"""
Test code for Heterodyne class.
"""

import os
import shutil
import subprocess as sp

import lal
import numpy as np
import pytest
from astropy.utils.data import download_file
from cwinpy import Heterodyne, HeterodynedData
from gwosc.api import DEFAULT_URL as GWOSC_DEFAULT_HOST
from lalpulsar.PulsarParametersWrapper import PulsarParametersPy
from lalpulsar.simulateHeterodynedCW import DOWNLOAD_URL, HeterodynedCWSimulator


class TestHeterodyne(object):
    """
    Test the Heterodyne object.
    """

    @classmethod
    def setup_class(cls):
        # create dummy frame cache files
        cls.dummydir = "testing_frame_cache"
        os.makedirs(cls.dummydir, exist_ok=True)
        cls.dummy_cache_files = []
        for i in range(0, 5):
            dummyfile = os.path.join(
                cls.dummydir, "frame_cache_{0:01d}.cache".format(i)
            )
            cls.dummy_cache_files.append(dummyfile)
            with open(dummyfile, "w") as fp:
                fp.write("blah\n")

        # create some fake data frames using lalapps_Makefakedata_v5
        mfd = shutil.which("lalapps_Makefakedata_v5")

        cls.fakedatadir = "testing_fake_frame_cache"
        cls.fakedatadetectors = ["H1", "L1"]
        cls.fakedatachannels = [
            "{}:FAKE_DATA".format(det) for det in cls.fakedatadetectors
        ]
        cls.fakedatastarts = [1000000000, 1000000000 + 86400 * 2]
        cls.fakedataduration = 86400

        os.makedirs(cls.fakedatadir, exist_ok=True)

        cls.fakedatabandwidth = 8  # Hz
        sqrtSn = 1e-29  # noise amplitude spectral density
        cls.fakedataname = "FAKEDATA"

        # Create two pulsars to inject: one isolated and one binary
        cls.fakepulsarpar = []

        # requirements for Makefakedata pulsar input files
        isolatedstr = """\
Alpha = {alpha}
Delta = {delta}
Freq = {f0}
f1dot = {f1}
f2dot = {f2}
refTime = {pepoch}
h0 = {h0}
cosi = {cosi}
psi = {psi}
phi0 = {phi0}
"""

        binarystr = """\
orbitasini = {asini}
orbitPeriod = {period}
orbitTp = {Tp}
orbitArgp = {argp}
orbitEcc = {ecc}
"""

        transientstr = """\
transientWindowType = {wintype}
transientStartTime = {tstart}
transientTau = {tau}
"""

        # FIRST PULSAR (ISOLATED)
        f0 = 6.9456 / 2.0  # source rotation frequency (Hz)
        f1 = -9.87654e-11 / 2.0  # source rotational frequency derivative (Hz/s)
        f2 = 2.34134e-18 / 2.0  # second frequency derivative (Hz/s^2)
        alpha = 0.0  # source right ascension (rads)
        delta = 0.5  # source declination (rads)
        pepoch = 1000000000  # frequency epoch (GPS)

        # GW parameters
        h0 = 3.0e-24  # GW amplitude
        phi0 = 1.0  # GW initial phase (rads)
        cosiota = 0.1  # cosine of inclination angle
        psi = 0.5  # GW polarisation angle (rads)

        mfddic = {
            "alpha": alpha,
            "delta": delta,
            "f0": 2 * f0,
            "f1": 2 * f1,
            "f2": 2 * f2,
            "pepoch": pepoch,
            "h0": h0,
            "cosi": cosiota,
            "psi": psi,
            "phi0": phi0,
        }

        cls.fakepulsarpar.append(PulsarParametersPy())
        cls.fakepulsarpar[0]["PSRJ"] = "J0000+0000"
        cls.fakepulsarpar[0]["H0"] = h0
        cls.fakepulsarpar[0]["PHI0"] = phi0 / 2.0
        cls.fakepulsarpar[0]["PSI"] = psi
        cls.fakepulsarpar[0]["COSIOTA"] = cosiota
        cls.fakepulsarpar[0]["F"] = [f0, f1, f2]
        cls.fakepulsarpar[0]["RAJ"] = alpha
        cls.fakepulsarpar[0]["DECJ"] = delta
        cls.fakepulsarpar[0]["PEPOCH"] = pepoch
        cls.fakepulsarpar[0]["EPHEM"] = "DE405"
        cls.fakepulsarpar[0]["UNITS"] = "TDB"

        cls.fakepardir = "testing_fake_par_dir"
        os.makedirs(cls.fakepardir, exist_ok=True)
        cls.fakeparfile = []
        cls.fakeparfile.append(os.path.join(cls.fakepardir, "J0000+0000.par"))
        cls.fakepulsarpar[0].pp_to_par(cls.fakeparfile[-1])

        injfile = os.path.join(cls.fakepardir, "inj.dat")
        with open(injfile, "w") as fp:
            fp.write("[Pulsar 1]\n")
            fp.write(isolatedstr.format(**mfddic))
            fp.write("\n")

        # SECOND PULSAR (BINARY SYSTEM)
        f0 = 3.8654321 / 2.0  # source rotation frequency (Hz)
        f1 = 9.87654e-13 / 2.0  # source rotational frequency derivative (Hz/s)
        f2 = -1.34134e-20 / 2.0  # second frequency derivative (Hz/s^2)
        alpha = 1.3  # source right ascension (rads)
        delta = -0.4  # source declination (rads)
        pepoch = 1000086400  # frequency epoch (GPS)

        # GW parameters
        h0 = 7.5e-25  # GW amplitude
        phi0 = 0.7  # GW initial phase (rads)
        cosiota = 0.6  # cosine of inclination angle
        psi = 1.1  # GW polarisation angle (rads)

        # binary parameters
        asini = 1.4  # projected semi-major axis (ls)
        period = 0.1 * 86400  # orbital period (s)
        Tp = 999992083  # time of periastron (GPS)
        argp = 0.0  # argument of perisatron (rad)
        ecc = 0.09  # the orbital eccentricity

        mfddic = {
            "alpha": alpha,
            "delta": delta,
            "f0": 2 * f0,
            "f1": 2 * f1,
            "f2": 2 * f2,
            "pepoch": pepoch,
            "h0": h0,
            "cosi": cosiota,
            "psi": psi,
            "phi0": phi0,
        }

        mfdbindic = {
            "asini": asini,
            "Tp": Tp,
            "period": period,
            "argp": argp,
            "ecc": ecc,
        }

        cls.fakepulsarpar.append(PulsarParametersPy())
        cls.fakepulsarpar[1]["PSRJ"] = "J1111+1111"
        cls.fakepulsarpar[1]["H0"] = h0
        cls.fakepulsarpar[1]["PHI0"] = phi0 / 2.0
        cls.fakepulsarpar[1]["PSI"] = psi
        cls.fakepulsarpar[1]["COSIOTA"] = cosiota
        cls.fakepulsarpar[1]["F"] = [f0, f1, f2]
        cls.fakepulsarpar[1]["RAJ"] = alpha
        cls.fakepulsarpar[1]["DECJ"] = delta
        cls.fakepulsarpar[1]["PEPOCH"] = pepoch
        cls.fakepulsarpar[1]["BINARY"] = "BT"
        cls.fakepulsarpar[1]["E"] = ecc
        cls.fakepulsarpar[1]["A1"] = asini
        cls.fakepulsarpar[1]["T0"] = Tp
        cls.fakepulsarpar[1]["OM"] = argp
        cls.fakepulsarpar[1]["PB"] = period
        cls.fakepulsarpar[1]["EPHEM"] = "DE405"
        cls.fakepulsarpar[1]["UNITS"] = "TDB"

        cls.fakeparfile.append(os.path.join(cls.fakepardir, "J1111+1111.par"))
        cls.fakepulsarpar[1].pp_to_par(cls.fakeparfile[-1])

        with open(injfile, "a") as fp:
            fp.write("[Pulsar 2]\n")
            fp.write(isolatedstr.format(**mfddic))
            fp.write(binarystr.format(**mfdbindic))
            fp.write("\n")

        # THIRD PULSAR (GLITCHING PULSAR)
        f0 = 5.3654321 / 2.0  # source rotation frequency (Hz)
        f1 = -4.57654e-10 / 2.0  # source rotational frequency derivative (Hz/s)
        f2 = 1.34134e-18 / 2.0  # second frequency derivative (Hz/s^2)
        alpha = 4.6  # source right ascension (rads)
        delta = -0.9  # source declination (rads)
        pepoch = 1000000000 + 1.5 * 86400  # frequency epoch (GPS)

        # glitch parameters
        df0 = 0.0001  # EM glitch frequency jump
        df1 = 1.2e-11  # EM glitch frequency derivative jump
        df2 = -4.5e-19  # EM glitch frequency second derivative jump
        dphi = 1.1  # EM glitch phase offset
        glepoch = pepoch  # glitch epoch

        # GW parameters
        h0 = 8.7e-25  # GW amplitude
        phi0 = 0.142  # GW initial phase (rads)
        cosiota = -0.3  # cosine of inclination angle
        psi = 0.52  # GW polarisation angle (rads)

        # binary parameters
        asini = 2.9  # projected semi-major axis (ls)
        period = 0.3 * 86400  # orbital period (s)
        Tp = 999995083  # time of periastron (GPS)
        argp = 0.5  # argument of perisatron (rad)
        ecc = 0.09  # the orbital eccentricity

        # for MFD I need to create this as two transient pulsars using a
        # rectangular window cutting before and after the glitch
        mfddic = {
            "alpha": alpha,
            "delta": delta,
            "f0": 2 * f0,
            "f1": 2 * f1,
            "f2": 2 * f2,
            "pepoch": pepoch,
            "h0": h0,
            "cosi": cosiota,
            "psi": psi,
            "phi0": phi0,
        }

        mfdbindic = {
            "asini": asini,
            "Tp": Tp,
            "period": period,
            "argp": argp,
            "ecc": ecc,
        }

        mfdtransientdic = {
            "wintype": "rect",
            "tstart": cls.fakedatastarts[0],
            "tau": 86400,
        }

        # signal before the glitch
        with open(injfile, "a") as fp:
            fp.write("[Pulsar 3]\n")
            fp.write(isolatedstr.format(**mfddic))
            fp.write(binarystr.format(**mfdbindic))
            fp.write(transientstr.format(**mfdtransientdic))
            fp.write("\n")

        mfddic["f0"] = 2 * (f0 + df0)
        mfddic["f1"] = 2 * (f1 + df1)
        mfddic["f2"] = 2 * (f2 + df2)
        mfddic["phi0"] = phi0 + 2 * dphi

        mfdtransientdic["tstart"] = cls.fakedatastarts[1]

        # signal after the glitch
        with open(injfile, "a") as fp:
            fp.write("[Pulsar 4]\n")
            fp.write(isolatedstr.format(**mfddic))
            fp.write(binarystr.format(**mfdbindic))
            fp.write(transientstr.format(**mfdtransientdic))

        cls.fakepulsarpar.append(PulsarParametersPy())
        cls.fakepulsarpar[2]["PSRJ"] = "J2222+2222"
        cls.fakepulsarpar[2]["H0"] = h0
        cls.fakepulsarpar[2]["PHI0"] = phi0 / 2.0
        cls.fakepulsarpar[2]["PSI"] = psi
        cls.fakepulsarpar[2]["COSIOTA"] = cosiota
        cls.fakepulsarpar[2]["F"] = [f0, f1, f2]
        cls.fakepulsarpar[2]["RAJ"] = alpha
        cls.fakepulsarpar[2]["DECJ"] = delta
        cls.fakepulsarpar[2]["PEPOCH"] = pepoch
        cls.fakepulsarpar[2]["BINARY"] = "BT"
        cls.fakepulsarpar[2]["E"] = ecc
        cls.fakepulsarpar[2]["A1"] = asini
        cls.fakepulsarpar[2]["T0"] = Tp
        cls.fakepulsarpar[2]["OM"] = argp
        cls.fakepulsarpar[2]["PB"] = period
        cls.fakepulsarpar[2]["EPHEM"] = "DE405"
        cls.fakepulsarpar[2]["UNITS"] = "TDB"
        cls.fakepulsarpar[2]["GLEP"] = [glepoch]
        cls.fakepulsarpar[2]["GLF0"] = [df0]
        cls.fakepulsarpar[2]["GLF1"] = [df1]
        cls.fakepulsarpar[2]["GLF2"] = [df2]
        cls.fakepulsarpar[2]["GLPH"] = [dphi / (2 * np.pi)]

        cls.fakeparfile.append(os.path.join(cls.fakepardir, "J2222+2222.par"))
        cls.fakepulsarpar[2].pp_to_par(cls.fakeparfile[-1])

        # set ephemeris files
        efile = download_file(
            DOWNLOAD_URL.format("earth00-40-DE405.dat.gz"), cache=True
        )
        sfile = download_file(DOWNLOAD_URL.format("sun00-40-DE405.dat.gz"), cache=True)

        for datastart in cls.fakedatastarts:
            for i in range(len(cls.fakedatachannels)):
                cmds = [
                    "-F",
                    cls.fakedatadir,
                    "--outFrChannels={}".format(cls.fakedatachannels[i]),
                    "-I",
                    cls.fakedatadetectors[i],
                    "--sqrtSX={0:.1e}".format(sqrtSn),
                    "-G",
                    str(datastart),
                    "--duration={}".format(cls.fakedataduration),
                    "--Band={}".format(cls.fakedatabandwidth),
                    "--fmin",
                    "0",
                    '--injectionSources="{}"'.format(injfile),
                    "--outLabel={}".format(cls.fakedataname),
                    '--ephemEarth="{}"'.format(efile),
                    '--ephemSun="{}"'.format(sfile),
                ]

                # run makefakedata
                sp.run([mfd] + cmds)

        # create dummy segment file
        cls.dummysegments = [(1000000000, 1000000600), (1000000800, 1000000900)]
        cls.dummysegmentfile = os.path.join(cls.fakedatadir, "fakesegments.txt")
        with open(cls.dummysegmentfile, "w") as fp:
            for segs in cls.dummysegments:
                fp.write("{} {}\n".format(segs[0], segs[1]))

    @classmethod
    def teardown_class(cls):
        """
        Remove test simulation directory.
        """

        shutil.rmtree(cls.dummydir)
        shutil.rmtree(cls.fakepardir)
        shutil.rmtree(cls.fakedatadir)

    def test_start_end(self):
        """
        Test for valid start and end times.
        """

        starttime = "blah"
        endtime = "blah"

        with pytest.raises(TypeError):
            Heterodyne(starttime, endtime)

        starttime = 100000000.0
        with pytest.raises(TypeError):
            Heterodyne(starttime, endtime)

        endtime = 100
        with pytest.raises(ValueError):
            Heterodyne(starttime, endtime)

        endtime = starttime + 86400
        het = Heterodyne(starttime, endtime)

        assert type(het.starttime) is int
        assert type(het.endtime) is int
        assert int(starttime) == het.starttime
        assert int(endtime) == het.endtime

        het = Heterodyne()
        assert het.starttime is None
        assert het.endtime is None

        # try setter
        het.starttime = 1000000001.1
        assert het.starttime == int(1000000001.1)

        # try setter
        het.endtime = 1000000002.1
        assert het.endtime == int(1000000002.1)

        # test the stride (default value)
        assert het.stride == 3600

        with pytest.raises(TypeError):
            het.stride = "kgsdg"

        with pytest.raises(TypeError):
            het.stride = 1.5

        with pytest.raises(ValueError):
            het.stride = 0

        stride = 1
        het.stride = stride
        assert het.stride == stride

    def test_detector(self):
        """
        Test for valid detector.
        """

        detector = "lhsdlgda"
        with pytest.raises(ValueError):
            Heterodyne(detector=detector)

        detector = 348.23
        with pytest.raises(TypeError):
            Heterodyne(detector=detector)

        het = Heterodyne()

        assert het.detector is None
        assert het.laldetector is None

        # try setter
        detector = "H1"
        het.detector = detector

        assert het.detector == detector
        assert type(het.laldetector) is lal.Detector
        assert het.laldetector.frDetector.prefix == detector

    def test_frametype(self):
        """
        Test for valid frame type.
        """

        frametype = 1.0
        with pytest.raises(TypeError):
            Heterodyne(frametype=frametype)

        het = Heterodyne()
        assert het.frametype is None

        frametype = "H1_R"
        het.frametype = frametype
        assert het.frametype == frametype

    def test_channel(self):
        """
        Test for valid channel name.
        """

        channel = 1
        with pytest.raises(TypeError):
            Heterodyne(channel=channel)

        channel = "lsgdkgks"
        with pytest.raises(ValueError):
            Heterodyne(channel=channel)

        het = Heterodyne()

        assert het.channel is None
        assert het.detector is None

        detector = "L1"
        channel = "{}:GWOSC-4KHZ_R1_STRAIN".format(detector)

        het.detector = "H1"
        with pytest.raises(ValueError):
            het.channel = channel

        het.detector = None
        het.channel = channel

        assert het.channel == channel
        assert het.detector == detector

    def test_framecache(self):
        """
        Test frame cache file setting.
        """

        with pytest.raises(TypeError):
            Heterodyne(framecache=1.2)

        with pytest.raises(ValueError):
            Heterodyne(framecache="lsgdfklg")

        with pytest.raises(TypeError):
            Heterodyne(framecache=[1, 2])

        het = Heterodyne(framecache=self.dummy_cache_files[0])
        assert het.framecache == self.dummy_cache_files[0]

        het.framecache = self.dummy_cache_files
        assert len(het.framecache) == len(self.dummy_cache_files)
        for i, df in enumerate(self.dummy_cache_files):
            assert df == het.framecache[i]

    def test_host(self):
        """
        Test host name.
        """

        with pytest.raises(TypeError):
            Heterodyne(host=1.2)

        with pytest.raises(RuntimeError):
            Heterodyne(host="+0--23oiyds")

        het = Heterodyne()

        assert het.host is None

        host = "https://www.google.com"
        het.host = host

        assert het.host == host

        host = "www.google.com"
        het.host = host
        assert het.host == host

    def test_get_frame_data(self):
        """
        Test reading of local frame data.
        """

        het = Heterodyne()

        with pytest.raises(ValueError):
            # no start or end time set
            het.get_frame_data()

        with pytest.raises(ValueError):
            # no start time set
            het.get_frame_data(endtime=1000000000)

        with pytest.raises(ValueError):
            # no channel set
            het.get_frame_data(starttime=1000000000, endtime=1000084600)

        with pytest.raises(IOError):
            # invalid file name
            het.get_frame_data(
                starttime=1000000000,
                endtime=1000084600,
                channel=self.fakedatachannels[0],
                framecache="jhsdklgdks.txt",
            )

        with pytest.raises(IOError):
            # cache file contains invalid frames
            het.get_frame_data(
                starttime=1000000000,
                endtime=1000000000 + 3 * 86400,
                framecache=self.dummy_cache_files[0],
                site="H1",
                channel=self.fakedatachannels[0],
            )

        # test reading files/generating a local cache list (all files)
        cachefile = os.path.join(self.fakedatadir, "frcache.txt")
        data = het.get_frame_data(
            starttime=1000000000,
            endtime=1000000000 + 86400,
            framecache=self.fakedatadir,
            site="H1",
            outputframecache=cachefile,
            channel=self.fakedatachannels[0],
        )

        assert int(data.t0.value) == self.fakedatastarts[0]
        assert data.dt.value == 1 / (2 * (self.fakedatabandwidth))

        with open(cachefile, "r") as fp:
            cachedata = [fl.strip() for fl in fp.readlines()]

        assert len(cachedata) == 1
        for i in range(len(cachedata)):
            assert (
                "{}-{}_{}-{}-{}.gwf".format(
                    self.fakedatadetectors[0][0],
                    self.fakedatadetectors[0],
                    self.fakedataname,
                    self.fakedatastarts[i],
                    self.fakedataduration,
                )
                == os.path.basename(cachedata[i])
            )

        # test reading files from cache file
        data = het.get_frame_data(
            starttime=1000000000,
            endtime=1000000000 + 86400,
            framecache=cachefile,
            site="H1",
            channel=self.fakedatachannels[0],
        )

        assert int(data.t0.value) == self.fakedatastarts[0]
        assert data.dt.value == 1 / (2 * (self.fakedatabandwidth))

        with pytest.raises(IOError):
            # try reading data outside of range
            het.get_frame_data(
                starttime=900000000,
                endtime=900000000 + 2 * 86400,
                framecache=cachefile,
                site="H1",
                channel=self.fakedatachannels[0],
            )

        with pytest.raises(IOError):
            # try reading data from the wrong channel
            het.get_frame_data(
                starttime=1000000000,
                endtime=1000000000 + 86400,
                framecache=cachefile,
                site="H1",
                channel=self.fakedatachannels[1],
            )

        del het
        del data

        # test reading from GWOSC
        het = Heterodyne()
        data = het.get_frame_data(
            starttime=1126259460, endtime=1126259464, host=GWOSC_DEFAULT_HOST, site="H1"
        )

        assert int(data.t0.value) == 1126259460
        assert data.dt.value == 1 / 4096
        assert len(data) == 16384

    @pytest.mark.disable_socket
    def test_get_frame_data_no_internet(self):
        # test None if not able to access GWOSC data
        het = Heterodyne()
        data = het.get_frame_data(
            site="H1", starttime=1126259460, endtime=1126259464, host=GWOSC_DEFAULT_HOST
        )

        assert data is None

    def test_get_segments(self):
        """
        Test reading segment list file.
        """

        het = Heterodyne()

        segments = het.get_segment_list(segmentfile=self.dummysegmentfile)

        assert len(segments) == len(self.dummysegments)
        for sega, segb in zip(segments, self.dummysegments):
            for sa, sb in zip(sega, segb):
                assert sa == sb

        het = Heterodyne(segmentlist=self.dummysegments)

        assert len(segments) == len(self.dummysegments)
        for sega, segb in zip(segments, self.dummysegments):
            for sa, sb in zip(sega, segb):
                assert sa == sb

        with pytest.raises(IOError):
            # no existent segment file
            Heterodyne(segmentlist="klsghdfkdgskd")

        with pytest.raises(TypeError):
            Heterodyne(outputsegmentlist=1)

        with pytest.raises(TypeError):
            Heterodyne(appendsegmentlist=1.1)

        with pytest.raises(TypeError):
            Heterodyne(includeflags=1)

        with pytest.raises(TypeError):
            Heterodyne(excludeflags=1)

        with pytest.raises(TypeError):
            Heterodyne(segmentserver=1)

        het = Heterodyne(
            starttime=900000000, endtime=910000000, segmentlist=self.dummysegmentfile
        )

        assert len(het.segments) == 0

        het = Heterodyne(
            starttime=1000000850, endtime=1000001000, segmentlist=self.dummysegmentfile
        )

        assert len(het.segments) == 1
        assert het.segments[0][0] == 1000000850
        assert het.segments[0][1] == 1000000900

        het = Heterodyne(
            starttime=1000000100, endtime=1000000700, segmentlist=self.dummysegmentfile
        )

        assert len(het.segments) == 1
        assert het.segments[0][0] == 1000000100
        assert het.segments[0][1] == 1000000600

    def test_set_pulsars(self, capsys):
        """
        Test setting of pulsar parameter files.
        """

        with pytest.raises(TypeError):
            Heterodyne(pulsarfiles=1.2)

        het = Heterodyne(pulsarfiles=self.fakepardir)

        assert sorted(list(het.pulsarfiles.keys())) == [
            "J0000+0000",
            "J1111+1111",
            "J2222+2222",
        ]
        assert sorted(list(het.pulsarfiles.values())) == sorted(
            [
                os.path.realpath(self.fakeparfile[0]),
                os.path.realpath(self.fakeparfile[1]),
                os.path.realpath(self.fakeparfile[2]),
            ]
        )
        assert sorted(het.pulsars) == ["J0000+0000", "J1111+1111", "J2222+2222"]

        het = Heterodyne(pulsarfiles=self.fakeparfile[0])
        assert het.pulsarfiles == {"J0000+0000": self.fakeparfile[0]}
        assert het.pulsars == ["J0000+0000"]

        het = Heterodyne(pulsarfiles=self.fakeparfile[1])
        assert het.pulsarfiles == {"J1111+1111": self.fakeparfile[1]}
        assert het.pulsars == ["J1111+1111"]

        het = Heterodyne(pulsarfiles=self.fakeparfile[2])
        assert het.pulsarfiles == {"J2222+2222": self.fakeparfile[2]}
        assert het.pulsars == ["J2222+2222"]

        het = Heterodyne(pulsarfiles=[self.fakeparfile[0]])

        assert het.pulsarfiles == {"J0000+0000": self.fakeparfile[0]}
        assert het.pulsars == ["J0000+0000"]

        het = Heterodyne(pulsarfiles=self.fakeparfile)

        assert sorted(list(het.pulsarfiles.keys())) == [
            "J0000+0000",
            "J1111+1111",
            "J2222+2222",
        ]
        assert list(het.pulsarfiles.values()) == [
            self.fakeparfile[0],
            self.fakeparfile[1],
            self.fakeparfile[2],
        ]
        assert sorted(het.pulsars) == ["J0000+0000", "J1111+1111", "J2222+2222"]

        with pytest.raises(TypeError):
            Heterodyne(pulsarfiles=self.fakeparfile, pulsars=3.4)

        het = Heterodyne(pulsarfiles=self.fakeparfile, pulsars="J0328+5323")
        captured = capsys.readouterr()

        assert len(het.pulsars) == 0
        assert (
            captured.out
            == "Pulsars '['J0328+5323']' not included as no parameter files have been given for them\n"
        )

        het = Heterodyne(pulsarfiles=[self.fakeparfile[0]], pulsars=["J0000+0000"])

        assert het.pulsarfiles == {"J0000+0000": self.fakeparfile[0]}
        assert het.pulsars == ["J0000+0000"]

        with pytest.raises(TypeError):
            het.pulsars = 453

        pulsarfiles = {}
        pulsarfiles["J0000+0000"] = "kgsdkgf"
        het = Heterodyne(pulsarfiles=pulsarfiles)
        captured = capsys.readouterr()
        assert len(het.pulsarfiles) == 0
        assert (
            captured.out
            == "Pulsar file 'kgsdkgf' could not be read. This pulsar will be ignored.\n"
        )

        pulsarfiles = {}
        pulsarfiles["J0000+0001"] = os.path.realpath(self.fakeparfile[0])
        het = Heterodyne(pulsarfiles=pulsarfiles)
        captured = capsys.readouterr()
        assert len(het.pulsarfiles) == 1
        assert (
            captured.out
            == "Inconsistent naming in pulsarfile dictionary. Using pulsar name 'J0000+0000' from parameter file\n"
        )

        pulsarfiles = {}
        pulsarfiles["J0000+0000"] = os.path.realpath(self.fakeparfile[0])
        het = Heterodyne(pulsarfiles=pulsarfiles)
        assert het.pulsarfiles == {"J0000+0000": os.path.realpath(self.fakeparfile[0])}
        assert het.pulsars == ["J0000+0000"]

    def test_crop(self):
        """
        Test setting of crop.
        """

        with pytest.raises(ValueError):
            Heterodyne(crop=-1)

        with pytest.raises(TypeError):
            Heterodyne(crop=0.5)

        crop = 50
        het = Heterodyne(crop=crop)

        assert het.crop == crop

    def test_heterodyne(self):
        """
        Test heterodyning on fake data.
        """

        segments = [
            (time, time + self.fakedataduration) for time in self.fakedatastarts
        ]

        het = Heterodyne(
            pulsarfiles=self.fakeparfile,
            pulsars=["J0000+0000"],
            segmentlist=segments,
            framecache=self.fakedatadir,
            channel=self.fakedatachannels[0],
        )

        with pytest.raises(TypeError):
            het.stride = "ksgdk"

        with pytest.raises(TypeError):
            het.stride = 4.5

        with pytest.raises(ValueError):
            het.stride = -1

        with pytest.raises(TypeError):
            het.filterknee = "lshdl"

        with pytest.raises(ValueError):
            het.filterknee = 0

        with pytest.raises(TypeError):
            het.freqfactor = "ldkme"

        with pytest.raises(ValueError):
            het.freqfactor = -2.3

        with pytest.raises(ValueError):
            # test that not setting an output gives a error
            het.heterodyne()

        # test setting an output directory
        outdir = os.path.join(self.fakedatadir, "heterodyne_output")
        het.outputfiles = outdir

        assert len(het.outputfiles) == 1
        assert list(het.outputfiles.keys()) == ["J0000+0000"]
        assert list(het.outputfiles.values()) == [os.path.join(outdir, het.label)]

        with pytest.raises(ValueError):
            # attempt to include glitch evolution without setting includessb to True
            het.heterodyne(includeglitch=True)

        # perform first stage heterodyne
        het = Heterodyne(
            starttime=segments[0][0],
            endtime=segments[-1][-1],
            pulsarfiles=self.fakeparfile,
            segmentlist=segments,
            framecache=self.fakedatadir,
            channel=self.fakedatachannels[0],
            freqfactor=2,
            stride=self.fakedataduration // 2,
            output=outdir,
            resamplerate=1,
        )

        het.heterodyne()

        labeldict = {
            "det": het.detector,
            "gpsstart": int(het.starttime),
            "gpsend": int(het.endtime),
            "freqfactor": int(het.freqfactor),
        }

        # expected length (after cropping)
        length = (
            het.resamplerate * np.diff(segments).sum() - 2 * len(segments) * het.crop
        )

        # expected start time (after cropping)
        t0 = segments[0][0] + het.crop + 0.5 / het.resamplerate

        # expected end time (after croppping)
        tend = segments[-1][-1] - het.crop - 0.5 / het.resamplerate

        # check output
        for psr in ["J0000+0000", "J1111+1111", "J2222+2222"]:
            assert os.path.isfile(het.outputfiles[psr].format(**labeldict, psr=psr))

            hetdata = HeterodynedData.read(
                het.outputfiles[psr].format(**labeldict, psr=psr)
            )

            assert len(hetdata) == length
            assert het.resamplerate == hetdata.dt.value
            assert t0 == hetdata.times.value[0]
            assert tend == hetdata.times.value[-1]
            assert het.detector == hetdata.detector

        # perform second stage of heterodyne
        with pytest.raises(TypeError):
            Heterodyne(heterodyneddata=0)

        fineoutdir = os.path.join(self.fakedatadir, "fine_heterodyne_output")

        # first heterodyne without SSB
        het2 = Heterodyne(
            detector=self.fakedatadetectors[0],
            heterodyneddata=outdir,  # pass previous output directory
            pulsarfiles=self.fakeparfile,
            freqfactor=2,
            resamplerate=1 / 60,
            includessb=False,
            output=fineoutdir,
            label="heterodyne_{psr}_{det}_{freqfactor}.hdf5",
        )
        het2.heterodyne()

        models = []
        for i, psr in enumerate(["J0000+0000", "J1111+1111", "J2222+2222"]):
            # load data
            hetdata = HeterodynedData.read(
                het2.outputfiles[psr].format(**labeldict, psr=psr)
            )

            assert het2.resamplerate == 1 / hetdata.dt.value
            assert len(hetdata) == int(length * het2.resamplerate)

            # set expected model
            sim = HeterodynedCWSimulator(
                hetdata.par,
                hetdata.detector,
                times=hetdata.times.value,
                earth_ephem=hetdata.ephemearth,
                sun_ephem=hetdata.ephemsun,
            )

            # due to how the HeterodynedCWSimulator works we need to set
            # updateglphase = True for the glitching signal to generate a
            # signal without the glitch phase included!
            models.append(
                sim.model(
                    usephase=True,
                    freqfactor=hetdata.freq_factor,
                    updateglphase=(True if psr == "J2222+2222" else False),
                )
            )

            # without inclusion of SSB model should not match
            assert np.any(np.abs(hetdata.data - models[i]) / np.abs(models[i]) > 5e-3)

        # now heterodyne with SSB
        del het2
        het2 = Heterodyne(
            detector=self.fakedatadetectors[0],
            heterodyneddata=outdir,  # pass previous output directory
            pulsarfiles=self.fakeparfile,
            freqfactor=2,
            resamplerate=1 / 60,
            includessb=True,
            output=fineoutdir,
            label="heterodyne_{psr}_{det}_{freqfactor}.hdf5",
        )
        het2.heterodyne()

        for i, psr in enumerate(["J0000+0000", "J1111+1111", "J2222+2222"]):
            # load data
            hetdata = HeterodynedData.read(
                het2.outputfiles[psr].format(**labeldict, psr=psr)
            )

            assert het2.resamplerate == 1 / hetdata.dt.value
            assert len(hetdata) == int(length * het2.resamplerate)

            if psr == "J0000+0000":  # isolated pulsar
                assert np.all(
                    np.abs(hetdata.data - models[i]) / np.abs(models[i]) < 5e-3
                )
            else:
                # without inclusion of BSB/glitch phase model should not match
                assert np.any(
                    np.abs(hetdata.data - models[i]) / np.abs(models[i]) > 5e-3
                )

        # now heterodyne with SSB and BSB
        del het2
        het2 = Heterodyne(
            detector=self.fakedatadetectors[0],
            heterodyneddata={
                psr: het.outputfiles[psr].format(**labeldict, psr=psr)
                for psr in ["J0000+0000", "J1111+1111", "J2222+2222"]
            },  # test using dictionary
            pulsarfiles=self.fakeparfile,
            freqfactor=2,
            resamplerate=1 / 60,
            includessb=True,
            includebsb=True,
            output=fineoutdir,
            label="heterodyne_{psr}_{det}_{freqfactor}.hdf5",
        )
        het2.heterodyne()

        for i, psr in enumerate(["J0000+0000", "J1111+1111", "J2222+2222"]):
            # load data
            hetdata = HeterodynedData.read(
                het2.outputfiles[psr].format(**labeldict, psr=psr)
            )

            assert het2.resamplerate == 1 / hetdata.dt.value
            assert len(hetdata) == int(length * het2.resamplerate)

            if psr in [
                "J0000+0000",
                "J1111+1111",
            ]:  # isolated and binary pulsar (non-glitching)
                assert np.all(
                    np.abs(hetdata.data - models[i]) / np.abs(models[i]) < 1e-2
                )
            else:
                # without inclusion glitch phase model should not match
                assert np.any(
                    np.abs(hetdata.data - models[i]) / np.abs(models[i]) > 1e-2
                )

        # now heterodyne with SSB, BSB and glitch phase
        del het2
        het2 = Heterodyne(
            detector=self.fakedatadetectors[0],
            heterodyneddata={
                psr: het.outputfiles[psr].format(**labeldict, psr=psr)
                for psr in ["J0000+0000", "J1111+1111", "J2222+2222"]
            },  # test using dictionary
            pulsarfiles=self.fakeparfile,
            freqfactor=2,
            resamplerate=1 / 60,
            includessb=True,
            includebsb=True,
            includeglitch=True,
            output=fineoutdir,
            label="heterodyne_{psr}_{det}_{freqfactor}.hdf5",
        )
        het2.heterodyne()

        for i, psr in enumerate(["J0000+0000", "J1111+1111", "J2222+2222"]):
            # load data
            hetdata = HeterodynedData.read(
                het2.outputfiles[psr].format(**labeldict, psr=psr)
            )

            assert het2.resamplerate == 1 / hetdata.dt.value
            assert len(hetdata) == int(length * het2.resamplerate)

            # increase tolerance for acceptance due to small outliers (still
            # equivalent at the ~2% level)
            assert np.all(np.abs(hetdata.data - models[i]) / np.abs(models[i]) < 2e-2)

    def test_full_heterodyne(self):
        """
        Test heterodyning on fake data, performing the heterodyne in one step.
        """

        segments = [
            (time, time + self.fakedataduration) for time in self.fakedatastarts
        ]

        # perform heterodyne in one step
        fulloutdir = os.path.join(self.fakedatadir, "full_heterodyne_output")

        het = Heterodyne(
            starttime=segments[0][0],
            endtime=segments[-1][-1],
            pulsarfiles=self.fakeparfile,
            segmentlist=segments,
            framecache=self.fakedatadir,
            channel=self.fakedatachannels[0],
            freqfactor=2,
            stride=self.fakedataduration // 2,
            resamplerate=1 / 60,
            includessb=True,
            includebsb=True,
            includeglitch=True,
            output=fulloutdir,
            label="heterodyne_{psr}_{det}_{freqfactor}.hdf5",
        )

        het.heterodyne()

        labeldict = {
            "det": het.detector,
            "gpsstart": int(het.starttime),
            "gpsend": int(het.endtime),
            "freqfactor": int(het.freqfactor),
        }

        # compare against model
        for i, psr in enumerate(["J0000+0000", "J1111+1111", "J2222+2222"]):
            # load data
            hetdata = HeterodynedData.read(
                het.outputfiles[psr].format(**labeldict, psr=psr)
            )

            assert het.resamplerate == 1 / hetdata.dt.value

            # set expected model
            sim = HeterodynedCWSimulator(
                hetdata.par,
                hetdata.detector,
                times=hetdata.times.value,
                earth_ephem=hetdata.ephemearth,
                sun_ephem=hetdata.ephemsun,
            )

            # due to how the HeterodynedCWSimulator works we need to set
            # updateglphase = True for the glitching signal to generate a
            # signal without the glitch phase included!
            model = sim.model(
                usephase=True,
                freqfactor=hetdata.freq_factor,
                updateglphase=(True if psr == "J2222+2222" else False),
            )

            # increase tolerance for acceptance due to small outliers (still
            # equivalent at the ~2% level)
            assert np.all(np.abs(hetdata.data - model) / np.abs(model) < 2e-2)



#!/usr/bin/env python

"""
Create HTCondor Dag to run a P-P test for the standard four parameters
of a pulsar gravitational-wave signal, assuming emission at twice the rotation
frequency: h0, iota, psi and phi0.
"""

import bilby
import numpy as np
from cwinpy.pe.testing import PEPPPlotsDAG

# set the priors
prior = {}
prior["h0"] = bilby.core.prior.Uniform(minimum=0.0, maximum=1e-22, latex_label="$h_0$")
prior["phi0"] = bilby.core.prior.Uniform(
    name="phi0", minimum=0.0, maximum=np.pi, latex_label=r"$\phi_0$", unit="rad"
)
prior["iota"] = bilby.core.prior.Sine(
    name="iota", minimum=0.0, maximum=np.pi, latex_label=r"$\iota$", unit="rad"
)
prior["psi"] = bilby.core.prior.Uniform(
    name="psi", minimum=0.0, maximum=np.pi / 2, latex_label=r"$\psi$", unit="rad"
)

# Maximum amplitude for any of the injection signal (below the prior maximum)
maxamp = 2.5e-24
det = "H1"  # detector and noise ASD
ninj = 250  # number of simulated signals
basedir = "/home/sismp2/pptest/fourparameters"  # base directory
accuser = "matthew.pitkin"
accgroup = "aluk.dev.o3.cw.targeted"
sampler = "dynesty"
numba = True
getenv = True
freqrange = (100.0, 200.0)

run = PEPPPlotsDAG(
    prior,
    ninj=ninj,
    maxamp=maxamp,
    detector=det,
    accountuser=accuser,
    accountgroup=accgroup,
    sampler=sampler,
    numba=numba,
    getenv=getenv,
    freqrange=freqrange,
    basedir=basedir,
    submit=True,  # submit the DAG
)



#!/usr/bin/env python

"""
An example estimating the posterior on h0 for some fake data in a targeted
pulsar-type search using the pe interface.
"""

import numpy as np
from bilby.core.prior import Uniform
from cwinpy.pe import pe
from matplotlib import pyplot as pl

# create a fake pulsar parameter file
parcontent = """\
PSRJ     J0123+3456
RAJ      01:23:45.6789
DECJ     34:56:54.321
F0       567.89
F1       -1.2e-12
PEPOCH   56789
H0       5.12e-25
COSIOTA  0.3
PSI      1.1
PHI0     2.4
"""

# add content to the par file
parfile = "J0123+3456.par"
with open(parfile, "w") as fp:
    fp.write(parcontent)

# use pe to create the data and sample on a grid
detector = "H1"  # the detector to use
times = np.linspace(1000000000.0, 1000086340.0, 1440)  # times
asd = 1e-24

# set prior on h0
priors = dict()
priors["h0"] = Uniform(0.0, 1e-24, "h0")
h0s = np.linspace(0.0, 1e-24, 500)  # h0 values to evaluate at

run = pe(
    detector=detector,
    fake_times=times,
    par_file=parfile,
    inj_par=parfile,
    fake_asd=asd,
    grid=True,
    grid_kwargs={"grid_size": {"h0": h0s}},
    prior=priors,
)

pl.plot(h0s, np.exp(run.grid.ln_posterior - np.max(run.grid.ln_posterior)), "b")
pl.axvline(run.hetdata["H1"][0].par["H0"])
pl.show()



"""
ASCII I/O registrations for cwinpy.HeterodynedData objects
"""

from ..data import HeterodynedData
from .readers import register_ascii_series_io

# -- registration -------------------------------------------------------------

register_ascii_series_io(HeterodynedData, format="txt")
register_ascii_series_io(HeterodynedData, format="txt.gz")



"""
Add _called_from_test if running code within as test, see
https://docs.pytest.org/en/latest/example/simple.html#detect-if-running-from-within-a-pytest-run
"""


def pytest_configure(config):
    import cwinpy

    cwinpy._called_from_test = True


def pytest_unconfigure(config):
    import cwinpy

    del cwinpy._called_from_test



# coding: utf-8

"""
A Python module for Bayesian inferences with continuous gravitational wave
sources.
"""

from setuptools import setup

import os

import versioneer


def readfile(filename):
    with open(filename, encoding="utf-8") as fp:
        filecontents = fp.read()
    return filecontents


VERSION = versioneer.get_version()
CMDCLASS = versioneer.get_cmdclass()

setup(
    name="cwinpy",
    version=VERSION,
    author="Matthew Pitkin",
    author_email="matthew.pitkin@ligo.org",
    packages=["cwinpy", "cwinpy.pe", "cwinpy.iostream"],
    url="http://git.ligo.org/CW/software/cwinpy",
    license="MIT",
    description="A Python module for Bayesian inferences with continuous gravitational-wave sources",
    long_description=readfile(os.path.join(os.path.dirname(__file__), "README.md")),
    long_description_content_type="text/markdown",
    install_requires=readfile(
        os.path.join(os.path.dirname(__file__), "requirements.txt")
    ),
    entry_points={
        "console_scripts": [
            "cwinpy_pe=cwinpy.pe.pe:pe_cli",
            "cwinpy_pe_dag=cwinpy.pe.pe:pe_dag_cli",
            "cwinpy_pe_generate_pp_plots=cwinpy.pe.testing:generate_pp_plots",
        ]
    },
    cmdclass=CMDCLASS,
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
)



"""
Run P-P plot testing for cwinpy_pe.
"""

import ast
import glob
import json
import os
import shutil
import sys
from argparse import ArgumentParser
from configparser import ConfigParser

import astropy.units as u
import bilby
import numpy as np
from astropy.coordinates import SkyCoord
from bilby_pipe.pp_test import read_in_result_list

from ..utils import int_to_alpha
from .pe import pe_dag


def generate_pp_plots(**kwargs):  # pragma: no cover
    """
    Script entry point, or function, to generate P-P plots (see, e.g., [1]_): a
    frequentist-style evaluation to test that the true value of a parameter
    falls with a given Bayesian credible interval the "correct" amount of
    times, provided the trues values are drawn from the same prior as used when
    evaluating the posteriors.

    Parameters
    ----------
    path: str
        Glob-able directory path pattern to a set of JSON format bilby results
        files.
    output: str
        Output filename for the PP plot.
    parameters: list
        A list of the parameters to include in the PP plot.
    snrs: bool
        Create a plot of the injected signal-to-noise ratio distribution.
        Defaults to False.

    References
    ----------

    .. [1] `T. Sidery et al. <https://arxiv.org/abs/1312.6013>`_,
       *PRD*, **89**, 084060 (2014)
    """

    if "path" in kwargs:
        path = kwargs["path"]

        # default output file is a PNG image in the current directory
        outfile = kwargs.get("output", os.path.join(os.getcwd(), "ppplot.png"))

        # default parameters are obtained from the results file prior
        parameters = kwargs.get("parameters", None)

        # create SNR plot
        snrs = kwargs.get("snrs", False)
    elif "cwinpy_pe_generate_pp_plots" == os.path.split(sys.argv[0])[-1]:
        parser = ArgumentParser(
            description=("A script to create a PP plot of CW signal parameters")
        )
        parser.add_argument(
            "--path",
            "-p",
            required=True,
            help=(
                "A glob-able path pattern to a set of bilby JSON results "
                "files for the set of simulations."
            ),
            dest="path",
        )
        parser.add_argument(
            "--output",
            "-o",
            help=("The output plot file name [default: %(default)s]."),
            default=os.path.join(os.getcwd(), "ppplot.png"),
            dest="outfile",
        )
        parser.add_argument(
            "--parameter",
            action="append",
            help=("The parameters with which to create the PP plots."),
        )
        parser.add_argument(
            "--snrs",
            action="store_true",
            default=False,
            help=("Set to plot distribution of injected signal-to-noise ratios."),
        )

        args = parser.parse_args()
        path = args.path
        outfile = args.outfile
        parameters = args.parameter
        snrs = args.snrs

        try:
            path = ast.literal_eval(path)
        except ValueError:
            pass
    else:
        raise KeyError("A 'path' keyword must be supplied.")

    # get results files
    try:
        resfiles = [
            rfile for rfile in glob.glob(path) if os.path.splitext(rfile)[1] == ".json"
        ]
    except Exception as e:
        raise IOError("Problem finding results files: {}".format(e))

    if len(resfiles) == 0:
        raise IOError("Problem finding results files. Probably an invalid path!")

    # read in results (need to create dummy class for args)
    class DummyClass(object):
        directory = path
        print = False

    results = read_in_result_list(DummyClass(), resfiles)

    if parameters is None:
        # get parameters to use from results file prior
        parameters = [
            name
            for name, p in results[0].priors.items()
            if isinstance(p, str) or p.is_fixed is False
        ]

    # make plots
    try:
        _ = bilby.core.result.make_pp_plot(results, filename=outfile, keys=parameters)
    except Exception as e:
        raise RuntimeError("Problem creating PP plots: {}".format(e))

    # make SNR plot
    if snrs:
        try:
            snrfiles = [
                snrfile
                for snrfile in glob.glob(path)
                if os.path.splitext(snrfile)[1] == ".snr"
            ]
        except Exception as e:
            raise IOError("Problem finding SNR files: {}".format(e))

        if len(snrfiles) == 0:
            raise IOError("Problem finding SNR files. Probably an invalid path!")

        snrvals = []
        for snrfile in snrfiles:
            with open(snrfile, "r") as fp:
                try:
                    snrvals.append(json.load(fp)["Injected SNR"])
                except Exception as e:
                    raise IOError(
                        "Problem loading SNR value from {}:\n{}".format(snrfile, e)
                    )

        from matplotlib import pyplot as pl

        fig, ax = pl.subplots(figsize=(10, 8))
        ax.hist(snrvals, bins=20, histtype="stepfilled")
        ax.set_xlabel("Injected signal-to-noise ratio", fontname="Gentium", fontsize=14)
        ax.set_ylabel("Number of sources", fontname="Gentium", fontsize=14)
        fig.tight_layout()

        try:
            fig.savefig(os.path.join(os.path.split(outfile)[0], "snrs.png"), dpi=250)
        except Exception as e:
            raise IOError("Problem saving figure: {}".format(e))


class PEPPPlotsDAG(object):
    """
    This class will generate a HTCondor Dagman job to create a number of
    simulated gravitational-wave signals from pulsars in Gaussian noise. These
    will be analysed using the ``cwinpy_pe`` script to sample the posterior
    probability distributions of the required parameter space. For each
    simulation and parameter combination the credible interval (bounded at the
    low end by the lowest sample in the posterior) in which the known true
    signal value lies will be found. The cumulative probability of finding the
    true value within a given credible interval is then plotted.

    Parameters
    ----------
    prior: dict
        A bilby-style prior dictionary giving the prior distributions from
        which to draw the injected signal values, and to use for signal
        recovery.
    ninj: int
        The number of simulated signals to create. Defaults to 100.
    maxamp: float
        A maximum on the amplitude parameter(s) to use when drawing the
        injection parameters. If none is given then this will be taken
        from the prior if using an amplitude parameter.
    basedir: str
        The base directory into which the simulations and outputs will be
        placed. If None then the current working directory will be used.
    detector: str, list
        A string, or list of strings, of detector prefixes for the simulated
        data. This defaults to a single detector - the LIGO Hanford
        Observatory - from which the simulated noise will be drawn from the
        advanced detector design sensitivity curve (e.g., [1]_).
    submit: bool
        Set whether to submit the Condor DAG or not.
    accountuser: str
        Value to give to the 'account_user' parameter in the Condor submit
        file. Default is to set no value.
    accountgroup: str
        Value to give to the 'account_user_group' parameter in the Condor
        submit file. Default is to set no value.
    getenv: bool
        Set the value for the 'getenv' parameter in the Condor submit file.
        Default is False.
    sampler: str
        The sampler to use. Defaults to dynesty.
    sampler_kwargs: dict
        A dictionary of keyword arguments for the sampler. Defaults to None.
    freqrange: list, tuple
        A pair of values giving the lower and upper rotation frequency ranges
        (in Hz) for the simulated signals. Defaults to (10, 750) Hz.
    outputsnr: bool
        Set whether to output the injected and recovered signal-to-noise
        ratios. Defaults to True.
    numba: bool
        Set whether or not to use the likelihood with numba enabled.

    References
    ----------

    .. [1] L. Barsotti, S. Gras, M. Evans, P. Fritschel,
       `LIGO T1800044-v5 <https://dcc.ligo.org/LIGO-T1800044/public>`_ (2018)

    """

    def __init__(
        self,
        prior,
        ninj=100,
        maxamp=None,
        basedir=None,
        detector="AH1",
        submit=False,
        accountuser=None,
        accountgroup=None,
        getenv=False,
        sampler="dynesty",
        sampler_kwargs=None,
        freqrange=(10.0, 750.0),
        outputsnr=True,
        numba=False,
    ):

        if isinstance(prior, dict):
            self.prior = bilby.core.prior.PriorDict(dictionary=prior)
        else:
            raise TypeError("Prior must be a dictionary-type object")

        if ninj < 1:
            raise ValueError("A positive number of injection must be given")
        self.ninj = int(ninj)

        # set maximum amplitude if given
        self.maxamp = None
        if isinstance(maxamp, float):
            if maxamp > 0.0:
                self.maxamp = maxamp
            else:
                raise ValueError("Maximum amplitude must be positive")

        if basedir is not None:
            self.basedir = basedir
            self.makedirs(basedir)
        else:
            self.basedir = os.getcwd()

        # build output directory structure
        self.detector = detector
        if isinstance(self.detector, str):
            self.detector = [self.detector]
        if not isinstance(self.detector, list):
            raise TypeError("Detector must be a string or list of strings")

        # posterior sample results directory
        self.resultsdir = os.path.join(self.basedir, "results")
        self.makedirs(self.resultsdir)

        # create pulsar parameter files
        self.create_pulsars(freqrange=freqrange)

        # create dag configuration file
        self.accountuser = accountuser
        self.accountgroup = accountgroup
        self.getenv = getenv
        self.submit = submit
        self.sampler = sampler
        self.sampler_kwargs = sampler_kwargs
        self.outputsnr = outputsnr
        self.numba = numba
        self.create_config()

        # create the DAG for cwinpy_pe jobs
        self.runner = pe_dag(config=self.config, build=False)

        # add PP plot creation DAG
        self.ppplots()

        if self.submit:
            self.runner.dag.submit_dag()

    def makedirs(self, dir):
        """
        Make a directory tree recursively.
        """

        try:
            os.makedirs(dir, exist_ok=True)
        except Exception as e:
            raise IOError("Could not create directory: {}\n{}".format(dir, e))

    def create_pulsars(self, freqrange):
        """
        Create the pulsar parameter files based on the samples from the priors.

        Parameters
        ----------
        freqrange: list, tuple
            A pair of values giving the lower and upper rotation frequency ranges
            (in Hz) for the simulated signals.
        """

        # pulsar parameter file directory
        self.pulsardir = os.path.join(self.basedir, "pulsars")
        self.makedirs(self.pulsardir)

        # "amplitude" parameters
        amppars = ["h0", "c21", "c22", "q22"]

        if not isinstance(freqrange, (list, tuple, np.ndarray)):
            raise TypeError("Frequency range must be a list or tuple")
        else:
            if len(freqrange) != 2:
                raise ValueError(
                    "Frequency range must contain an upper and lower value"
                )

        self.pulsars = {}
        for i in range(self.ninj):
            pulsar = {}

            for param in self.prior:
                if self.maxamp is not None and param in amppars:
                    # set maximum amplitude if given
                    pulsar[param.upper()] = bilby.core.prior.Uniform(
                        name=param, minimum=0.0, maximum=self.maxamp
                    ).sample()
                else:
                    pulsar[param.upper()] = self.prior[param].sample()

            # draw sky position uniformly from the sky if no prior is given
            if "ra" not in self.prior:
                raval = np.random.uniform(0.0, 2.0 * np.pi)
            else:
                raval = pulsar.pop("ra")

            if "dec" not in self.prior:
                decval = -(np.pi / 2.0) + np.arccos(np.random.uniform(-1.0, 1.0))
            else:
                decval = pulsar.pop("dec")

            skypos = SkyCoord(raval * u.rad, decval * u.rad)
            pulsar["RAJ"] = skypos.ra.to_string(u.hour, fields=3, sep=":", pad=True)
            pulsar["DECJ"] = skypos.dec.to_string(u.deg, fields=3, sep=":", pad=True)

            # set (rotation) frequency upper and lower bounds
            if "f0" not in self.prior:
                pulsar["F0"] = np.random.uniform(freqrange[0], freqrange[1])

            # set pulsar name from sky position
            rastr = skypos.ra.to_string(u.hour, fields=2, sep="", pad=True)
            decstr = skypos.dec.to_string(
                u.deg, fields=2, sep="", pad=True, alwayssign=True
            )
            pname = "J{}{}".format(rastr, decstr)
            pnameorig = str(pname)  # copy of original name
            counter = 1
            while pname in self.pulsars:
                anum = int_to_alpha(counter)
                pname = pnameorig + anum
                counter += 1

            pulsar["PSRJ"] = pname

            # output file name
            pfile = os.path.join(self.pulsardir, "{}.par".format(pname))

            with open(pfile, "w") as fp:
                for param in pulsar:
                    fp.write("{}\t{}\n".format(param, pulsar[param]))

            self.pulsars[pname] = {}
            self.pulsars[pname]["file"] = pfile
            self.pulsars[pname]["parameters"] = pulsar

    def create_config(self):
        """
        Create the configuration parser for the DAG.
        """

        self.config = ConfigParser()

        self.config["run"] = {"basedir": self.basedir}

        self.config["dag"] = {"build": False}

        self.config["job"] = {}
        self.config["job"]["getenv"] = str(self.getenv)

        if self.accountgroup is not None:
            self.config["job"]["accounting_group"] = self.accountgroup
        if self.accountuser is not None:
            self.config["job"]["accounting_group_user"] = self.accountuser

        self.config["pe"] = {}
        self.config["pe"]["pulsars"] = self.pulsardir
        self.config["pe"]["injections"] = self.pulsardir
        self.config["pe"]["results"] = self.resultsdir
        self.config["pe"]["numba"] = str(self.numba)

        # set fake data
        if "h0" in self.prior or "c22" in self.prior or "q22" in self.prior:
            self.config["pe"]["fake-asd-2f"] = str(self.detector)
        if "c21" in self.prior and "c22" in self.prior:
            self.config["pe"]["fake-asd-1f"] = str(self.detector)
        if "c21" in self.prior and "c22 not in self.prior":
            self.config["pe"]["fake-asd-1f"] = str(self.detector)

        # set the prior file
        label = "ppplot"
        self.priorfile = os.path.join(self.basedir, "{}.prior".format(label))
        self.prior.to_file(outdir=self.basedir, label=label)

        self.config["pe"]["priors"] = self.priorfile
        self.config["pe"]["sampler"] = self.sampler
        if isinstance(self.sampler_kwargs, dict):
            self.config["pe"]["sampler_kwargs"] = str(self.sampler_kwargs)
        self.config["pe"]["output_snr"] = str(self.outputsnr)

    def ppplots(self):
        """
        Set up job to create PP plots.
        """

        from pycondor import Job

        # get executable
        jobexec = shutil.which("cwinpy_pe_generate_pp_plots")

        extra_lines = []
        if self.accountgroup is not None:
            extra_lines.append("accounting_group = {}".format(self.accountgroup))
        if self.accountuser is not None:
            extra_lines.append("accounting_group_user = {}".format(self.accountuser))

        # create cwinpy_pe Job
        job = Job(
            "cwinpy_pe_pp_plots",
            jobexec,
            error=self.runner.dag.inputs.pe_log_directory,
            log=self.runner.dag.inputs.pe_log_directory,
            output=self.runner.dag.inputs.pe_log_directory,
            submit=self.runner.dag.inputs.submit_directory,
            universe="vanilla",
            request_memory=self.runner.dag.inputs.request_memory,
            getenv=self.getenv,
            queue=1,
            requirements=self.runner.dag.inputs.requirements,
            retry=self.runner.dag.inputs.retry,
            extra_lines=extra_lines,
            dag=self.runner.dag.pycondor_dag,
        )

        jobargs = "--path '{}' ".format(os.path.join(self.basedir, "results", "*", "*"))
        jobargs += "--output {} ".format(os.path.join(self.basedir, "ppplot.png"))
        if self.outputsnr:
            jobargs += "--snrs "
        job.add_arg(jobargs)

        job.add_parents(
            self.runner.dag.pycondor_dag.nodes[:-1]
        )  # exclude cwinpy_pe_pp_plots job itself
        self.runner.dag.build()



"""
Classes for hierarchical parameter inference.
"""

from itertools import compress

import bilby
import numpy as np
from lintegrate import logtrapz
from scipy.interpolate import splev, splrep
from scipy.stats import expon, gaussian_kde, truncnorm

from .utils import ellipticity_to_q22, q22_to_ellipticity

#: Allowed distributions and their required hyperparameters
DISTRIBUTION_REQUIREMENTS = {
    "exponential": ["mu"],
    "gaussian": ["mu", "sigma", "weight"],
    "deltafunction": ["peak"],
    "powerlaw": ["alpha", "minimum", "maximum"],
}


class BaseDistribution(object):
    """
    The base class for the distribution, as defined by a set of
    hyperparameters, that you want to fit.

    Parameters
    ----------
    name: str
        The parameter for which this distribution is the prior.
    disttype: str
        The type of distribution, e.g., 'exponential', 'gaussian'.
    hyperparameters: dict
        A dictionary of hyperparameters for the distribution with the keys
        giving the parameter names, and values giving their fixed value, or
        a :class:`bilby.core.prior.Prior` for values that are to be inferred.
    low: float
        The lower bound of the distribution
    high: float
        The upper bound of the distribution
    """

    def __init__(self, name, disttype, hyperparameters={}, low=-np.inf, high=np.inf):
        self.name = name  # the parameter name
        self.disttype = disttype
        self.hyperparameters = hyperparameters
        self.low = low
        self.high = high

        if self.low >= self.high:
            raise ValueError("Lower bound is higher than upper bound!")

    @property
    def disttype(self):
        return self._disttype

    @disttype.setter
    def disttype(self, disttype):
        if disttype.lower() not in DISTRIBUTION_REQUIREMENTS.keys():
            raise ValueError('Distribution name "{}" is not ' "known".format(disttype))
        else:
            self._disttype = disttype.lower()

    @property
    def hyperparameters(self):
        return self._hyperparameters

    @hyperparameters.setter
    def hyperparameters(self, hyperparameters):
        if isinstance(hyperparameters, dict):
            # check is contains the required parameter names
            for key in hyperparameters.keys():
                if key.lower() not in DISTRIBUTION_REQUIREMENTS[self.disttype]:
                    raise KeyError(
                        'Unknown parameter "{}" for distribution '
                        '"{}"'.format(key, self.disttype)
                    )
            self._hyperparameters = {
                key.lower(): value for key, value in hyperparameters.items()
            }
        else:
            raise TypeError("hyperparameters must be a dictionary")

        # set fixed values
        self.fixed = self._hyperparameters

    @property
    def parameters(self):
        return list(self.hyperparameters.keys())

    @property
    def values(self):
        return list(self.hyperparameters.values())

    @property
    def unpacked_parameters(self):
        params = []
        for key, value in self.hyperparameters.items():
            if isinstance(value, (list, np.ndarray)):
                for i in range(len(value)):
                    params.append("{0}{1:d}".format(key, i))
            else:
                params.append(key)
        return params

    @property
    def unpacked_values(self):
        values = []
        for key, value in self.hyperparameters.items():
            if isinstance(value, (list, np.ndarray)):
                for i in range(len(value)):
                    values.append(value[i])
            else:
                values.append(value)
        return values

    def __getitem__(self, item):
        if item.lower() in self.parameters:
            return self.hyperparameters[item.lower()]
        elif item.lower() in self.unpacked_parameters:
            return self.unpacked_values[self.unpacked_parameters.index(item.lower())]
        elif item.lower() in DISTRIBUTION_REQUIREMENTS[self.disttype]:
            return None
        else:
            raise KeyError('"{}" is not a parameter in this distribution'.format(item))

    def __setitem__(self, item, value):
        if item.lower() not in self.hyperparameters.keys():
            if item.lower() in DISTRIBUTION_REQUIREMENTS[self.disttype]:
                self._hyperparameters[item.lower()] = value
            else:
                raise KeyError(
                    '"{}" is not a parameter in this distribution'.format(item)
                )
        else:
            self._hyperparameters[item.lower()] = value

    @property
    def fixed(self):
        """
        Return a dictionary keyed to parameter names and with boolean values
        indicating whether the parameter is fixed (True), or to be inferred
        (False).
        """

        return self._fixed

    @fixed.setter
    def fixed(self, hyperparameters):
        self._fixed = dict()

        for param, value in hyperparameters.items():
            if isinstance(value, (bilby.core.prior.Prior, bilby.core.prior.PriorDict)):
                self._fixed[param] = False
            elif isinstance(value, (list, np.ndarray)):
                self._fixed[param] = []
                for i in range(len(value)):
                    if isinstance(
                        value[i], (bilby.core.prior.Prior, bilby.core.prior.PriorDict)
                    ):
                        self._fixed[param].append(False)
                    elif isinstance(value[i], (int, float)):
                        self._fixed[param].append(True)
                    else:
                        raise TypeError("Hyperparameter type is not valid")
            elif isinstance(value, (int, float)):
                self._fixed[param] = True
            else:
                raise TypeError("Hyperparameter type is not valid")

    @property
    def unpacked_fixed(self):
        """
        Return a flattened version of ``fixed``, with multivalued parameters
        indexed.
        """

        fixed = dict()

        for param, value in zip(self.unpacked_parameters, self.unpacked_values):
            if isinstance(value, (bilby.core.prior.Prior, bilby.core.prior.PriorDict)):
                fixed[param] = False
            elif isinstance(value, (int, float)):
                fixed[param] = True
            else:
                raise TypeError("Hyperparameter type is not valid")

        return fixed

    @property
    def unknown_parameters(self):
        """
        A list of the parameters that are to be inferred.
        """

        return list(
            compress(
                self.unpacked_parameters, ~np.array(list(self.unpacked_fixed.values()))
            )
        )

    @property
    def unknown_priors(self):
        """
        A list of the :class:`~bilby.core.prior.Prior` for the parameters
        that are to be inferred.
        """

        return list(
            compress(
                self.unpacked_values, ~np.array(list(self.unpacked_fixed.values()))
            )
        )

    def log_pdf(self, value, hyperparameters):
        """
        The natural logarithm of the distribution's probability density
        function at the given value.

        Parameters
        ----------
        value: float
            The value at which to evaluate the probability.
        hyperparameters: dict
            A dictionary of the hyperparameter values that define the current
            state of the distribution.

        Returns
        -------
        lnpdf:
            The natural logarithm of the probability.
        """

        return np.nan

    def pdf(self, value, hyperparameters):
        """
        The distribution's probability density function at the given value.

        Parameters
        ----------
        value: float
            The value at which to evaluate the probability.
        hyperparameters: dict
            A dictionary of the hyperparameter values that define the current
            state of the distribution.

        Returns
        -------
        pdf:
            The probability density.
        """

        return np.exp(self.log_pdf(value, hyperparameters))

    def sample(self, hyperparameters, size=1):
        """
        Draw a sample from the distribution as defined by the given
        hyperparameters.

        Parameters
        ----------
        hyperparameters: dict
            A dictionary of the hyperparameter values that define the current
            state of the distribution.
        size: int
            The number of samples to draw from the distribution.

        Returns
        -------
        sample:
            A sample, or set of samples, from the distribution.
        """

        return None


class BoundedGaussianDistribution(BaseDistribution):
    """
    A distribution to define estimating the parameters of a (potentially
    multi-modal) bounded Gaussian distribution.

    An example of using this distribution for a two component Gaussian
    distribution bounded at zero and with unknown mean, standard deviations and
    weights would be:

    >>> from bilby.core.prior import HalfNormal, LogUniform, DirichletPriorDict
    >>> # set priors for means (half-Normal distributions with mode at 0)
    >>> mus = [HalfNormal(10.0, name="mu0"), HalfNormal(10.0, name="mu1")]
    >>> # set priors for standard deviations (log uniform distributions)
    >>> sigmas = [LogUniform(name="sigma0", minimum=0.0001, maximum=100.0),
                  LogUniform(name="sigma1", minimum=0.0001, maximum=100.0)]
    >>> # set a Dirichlet prior on the weights (i.e., they must add up to 1)
    >>> weights = DirichletPriorDict(n_dim=2, label="weight")
    >>> dist = BoundedGaussianDistribution("x", mus=mus, sigmas=sigmas, weights=weights)

    Parameters
    ----------
    name: str
        See :class:`~cwinpy.hierarchical.BaseDistribution`
    mus: array_like
        A list of values of the means of each mode of the Gaussian.
    sigmas: array_like
        A list of values of the standard deviations of each mode of the
        Gaussian.
    weights: array_like
        A list of values of the weights (relative probabilities) of
        each mode. This will default to equal weights if not given. If wanting
        to estimate multiple weights a DirichletPriorDict should be used as in
        the example above.
    low: float
        The lower bound of the distribution (defaults to 0, i.e., only positive
        values are allowed)
    high: float
        The upper bound of the distribution (default to infinity)

    """

    def __init__(self, name, mus=[], sigmas=[], weights=None, low=0.0, high=np.inf):
        gaussianparameters = {"mu": [], "sigma": [], "weight": []}

        if isinstance(mus, (int, float, bilby.core.prior.Prior)):
            mus = [mus]
        elif not isinstance(mus, (list, np.ndarray)):
            raise TypeError("Unknown type for 'mus'")

        if isinstance(sigmas, (int, float, bilby.core.prior.Prior)):
            sigmas = [sigmas]
        elif not isinstance(sigmas, (list, np.ndarray)):
            raise TypeError("Unknown type for 'sigmas'")

        if weights is None:
            weights = [1] * len(mus)
        elif not isinstance(
            weights, (list, np.ndarray, bilby.core.prior.DirichletPriorDict)
        ):
            raise TypeError("Unknown type for 'weights'")

        if isinstance(weights, bilby.core.prior.DirichletPriorDict):
            # DirichletPriorDict has length one less than the number of weights
            nweights = len(weights) + 1
            gaussianparameters["weight"] = weights
        else:
            nweights = len(weights)

        # set the number of modes
        self.nmodes = len(mus)

        if len(mus) != len(sigmas) or nweights != len(mus):
            raise ValueError("'mus', 'sigmas' and 'weights' must be the same length")

        if self.nmodes < 1:
            raise ValueError("Gaussian must have at least one mode")

        for i in range(self.nmodes):
            gaussianparameters["mu"].append(mus[i])
            gaussianparameters["sigma"].append(sigmas[i])

            if isinstance(weights, (list, np.ndarray)):
                gaussianparameters["weight"].append(weights[i])

        # initialise
        super().__init__(
            name, "gaussian", hyperparameters=gaussianparameters, low=low, high=high
        )

    def log_pdf(self, value, hyperparameters={}):
        """
        The natural logarithm of the pdf of a 1d (potentially multi-modal)
        Gaussian probability distribution.

        Parameters
        ----------
        value: float
            The value at which the probability is to be evaluated.
        hyperparameters: dict
            A dictionary containing the current values of the hyperparameters
            that need to be inferred. If there are multiple modes and weights
            are not fixed then the hyperparameters should include ``n-1``
            weights values, where ``n`` is the number of modes.

        Returns
        -------
        logpdf:
            The natural logarithm of the probability density at the given
            value.
        """

        if np.any((value < self.low) | (value > self.high)):
            return -np.inf

        mus = self["mu"]
        sigmas = self["sigma"]

        if isinstance(self.fixed["weight"], (list, np.ndarray)):
            weights = self["weight"]
        else:
            weights = np.zeros(self.nmodes)

        # get current mus and sigmas from values
        for i in range(self.nmodes):
            if not self.fixed["mu"][i]:
                param = "mu{}".format(i)
                try:
                    mus[i] = hyperparameters[param]
                except KeyError:
                    raise KeyError(
                        "Cannot calculate log probability when "
                        "value '{}' is not given".format(param)
                    )

            if not self.fixed["sigma"][i]:
                param = "sigma{}".format(i)
                try:
                    sigmas[i] = hyperparameters[param]
                except KeyError:
                    raise KeyError(
                        "Cannot calculate log probability when "
                        "value '{}' is not given".format(param)
                    )

            if not isinstance(self.fixed["weight"], (list, np.ndarray)):
                param = "weight{}".format(i)
                if i < (self.nmodes - 1):
                    try:
                        weights[i] = hyperparameters[param]
                    except KeyError:
                        raise KeyError(
                            "Cannot calculate log probability when "
                            "value '{}' is not given".format(param)
                        )
                else:
                    # set final weight
                    weights[i] = 1.0 - np.sum(weights[:-1])

        if np.any(np.asarray(sigmas) <= 0.0):
            return -np.inf

        if np.any(np.asarray(weights) <= 0.0):
            return -np.inf

        # normalise weights
        lweights = np.log(np.asarray(weights) / np.sum(weights))

        # get log pdf
        if isinstance(value, (float, int)):
            logpdf = -np.inf
        elif isinstance(value, (list, np.ndarray)):
            logpdf = np.full_like(value, -np.inf)
        else:
            raise TypeError("value must be a float or array-like")

        for mu, sigma, lweight in zip(mus, sigmas, lweights):
            lpdf = lweight + truncnorm.logpdf(
                value,
                (self.low - mu) / sigma,
                (self.high - mu) / sigma,
                loc=mu,
                scale=sigma,
            )
            logpdf = np.logaddexp(logpdf, lpdf)

        return logpdf

    def sample(self, hyperparameters={}, size=1):
        """
        Draw a sample from the bounded Gaussian distribution as defined by the
        given hyperparameters.

        Parameters
        ----------
        hyperparameters: dict
            A dictionary of the hyperparameter values that define the current
            state of the distribution. If there are multiple modes and weights
            are not fixed then the hyperparameters should include ``n-1``
            weights values, where ``n`` is the number of modes.
        size: int
            The number of samples to draw. Default is 1.

        Returns
        -------
        sample:
            A sample, or set of samples, from the distribution.
        """

        mus = self["mu"]
        sigmas = self["sigma"]

        if isinstance(self.fixed["weight"], (list, np.ndarray)):
            weights = self["weight"]
        else:
            weights = np.zeros(self.nmodes)

        # get current mus and sigmas from values
        for i in range(self.nmodes):
            if not self.fixed["mu"][i]:
                param = "mu{}".format(i)
                try:
                    mus[i] = hyperparameters[param]
                except KeyError:
                    raise KeyError(
                        "Cannot calculate log probability when "
                        "value '{}' is not given".format(param)
                    )

            if not self.fixed["sigma"][i]:
                param = "sigma{}".format(i)
                try:
                    sigmas[i] = hyperparameters[param]
                except KeyError:
                    raise KeyError(
                        "Cannot calculate log probability when "
                        "value '{}' is not given".format(param)
                    )

            if not isinstance(self.fixed["weight"], (list, np.ndarray)):
                param = "weight{}".format(i)
                if i < (self.nmodes - 1):
                    try:
                        weights[i] = hyperparameters[param]
                    except KeyError:
                        raise KeyError(
                            "Cannot calculate log probability when "
                            "value '{}' is not given".format(param)
                        )
                else:
                    # set final weight
                    weights[i] = 1.0 - np.sum(weights[:-1])

        # cumulative normalised weights
        cweights = np.cumsum(np.asarray(weights) / np.sum(weights))

        # pick mode and draw sample
        if self.nmodes == 1:
            sample = truncnorm.rvs(
                (self.low - mus[0]) / sigmas[0],
                (self.high - mus[0]) / sigmas[0],
                loc=mus[0],
                scale=sigmas[0],
                size=size,
            )
        else:
            sample = np.zeros(size)
            for i in range(size):
                mode = np.argwhere(cweights - np.random.rand() > 0)[0][0]

                sample[i] = truncnorm.rvs(
                    (self.low - mus[mode]) / sigmas[mode],
                    (self.high - mus[mode]) / sigmas[mode],
                    loc=mus[mode],
                    scale=sigmas[mode],
                    size=1,
                )

            if size == 1:
                sample = sample[0]

        return sample


class ExponentialDistribution(BaseDistribution):
    """
    A distribution to define estimating the parameters of an exponential distribution.

    Parameters
    ----------
    name: str
        See :class:`~cwinpy.hierarchical.BaseDistribution`
    mu: float, Prior
        The mean of the exponential distribution.
    """

    def __init__(self, name, mu):
        # initialise
        super().__init__(
            name, "exponential", hyperparameters=dict(mu=mu), low=0.0, high=np.inf
        )

    def log_pdf(self, value, hyperparameters={}):
        """
        The natural logarithm of the pdf of an exponential distribution.

        Parameters
        ----------
        value: float
            The value at which the probability is to be evaluated.
        hyperparameters: dict
            A dictionary containing the current values of the hyperparameters
            that need to be inferred.

        Returns
        -------
        logpdf:
            The natural logarithm of the probability at the given value.
        """

        if np.any((value < self.low) | (value > self.high)):
            return -np.inf

        mu = self["mu"]
        if not self.fixed["mu"]:
            try:
                mu = hyperparameters["mu"]
            except KeyError:
                raise KeyError("Cannot evaluate the probability when mu is not given")

        if mu <= 0.0:
            return -np.inf

        # get log pdf
        logpdf = expon.logpdf(value, scale=mu)

        return logpdf

    def sample(self, hyperparameters={}, size=1):
        """
        Draw a sample from the exponential distribution as defined by the
        given hyperparameters.

        Parameters
        ----------
        hyperparameters: dict
            A dictionary of the hyperparameter values (``mu``) that define the
            current state of the distribution.
        size: int
            The number of samples to draw from the distribution.

        Returns
        -------
        sample:
            A sample, or set of samples, from the distribution.
        """

        mu = self["mu"]
        if not self.fixed["mu"]:
            try:
                mu = hyperparameters["mu"]
            except KeyError:
                raise KeyError("Cannot evaluate the probability when mu is not given")

        samples = expon.rvs(scale=mu, size=size)

        while 1:
            idx = (samples > self.low) & (samples < self.high)
            nvalid = np.sum(idx)

            if nvalid != size:
                sample = expon.rvs(scale=mu, size=(size - nvalid))
                samples[~idx] = sample
            else:
                break

        if size == 1:
            sample = samples[0]
        else:
            sample = samples

        return sample


class PowerLawDistribution(BaseDistribution):
    """
    A distribution to define estimating the parameters of a power law distribution.

    Parameters
    ----------
    name: str
        See :class:`~cwinpy.hierarchical.BaseDistribution`
    alpha: float, Prior
        The power law index of the distribution.
    minimum: float
        A positive finite value giving the lower cutoff of the distribution.
    maximum: float
        A positive finite value giving the upper cutoff of the distribution.
    """

    def __init__(self, name, alpha, minimum, maximum):
        if isinstance(minimum, float):
            if minimum <= 0 or not np.isfinite(minimum):
                raise ValueError(
                    "Minimum of distribution must be positive finite value"
                )

            if isinstance(maximum, float):
                if maximum < minimum:
                    raise ValueError(
                        "Maximum of distribution must be smaller than minimum!"
                    )

        if isinstance(maximum, float):
            if maximum <= 0 or not np.isfinite(maximum):
                raise ValueError(
                    "Maximum of distribution must be positive finite value"
                )

        # initialise
        super().__init__(
            name,
            "powerlaw",
            hyperparameters=dict(alpha=alpha, minimum=minimum, maximum=maximum),
        )

    def log_pdf(self, value, hyperparameters={}):
        """
        The natural logarithm of the pdf of a power law distribution.

        Parameters
        ----------
        value: float
            The value at which the probability is to be evaluated.
        hyperparameters: dict
            A dictionary containing the current values of the hyperparameters
            that need to be inferred.

        Returns
        -------
        logpdf:
            The natural logarithm of the probability at the given value.
        """

        alpha = self["alpha"]
        if not self.fixed["alpha"]:
            try:
                alpha = hyperparameters["alpha"]
            except KeyError:
                raise KeyError(
                    "Cannot evaluate the probability when alpha is not given"
                )

        minimum = self["minimum"]
        if not self.fixed["minimum"]:
            try:
                minimum = hyperparameters["minimum"]
            except KeyError:
                raise KeyError(
                    "Cannot evaluate the probability when minimum is not given"
                )
        elif np.any(value < minimum):
            return -np.inf

        maximum = self["maximum"]
        if not self.fixed["maximum"]:
            try:
                maximum = hyperparameters["maximum"]
            except KeyError:
                raise KeyError(
                    "Cannot evaluate the probability when maximum is not given"
                )
        elif np.any(value > maximum):
            return -np.inf

        # get log pdf
        logpdf = bilby.core.prior.PowerLaw(alpha, minimum, maximum).ln_prob(value)

        return logpdf

    def sample(self, hyperparameters={}, size=1):
        """
        Draw a sample from the exponential distribution as defined by the
        given hyperparameters.

        Parameters
        ----------
        hyperparameters: dict
            A dictionary of the hyperparameter values (``alpha``, ``minimum``
            and ``maximum``) that define the current state of the distribution.
        size: int
            The number of samples to draw from the distribution.

        Returns
        -------
        sample:
            A sample, or set of samples, from the distribution.
        """

        alpha = self["alpha"]
        if not self.fixed["alpha"]:
            try:
                alpha = hyperparameters["alpha"]
            except KeyError:
                raise KeyError(
                    "Cannot evaluate the probability when alpha is not given"
                )

        minimum = self["minimum"]
        if not self.fixed["minimum"]:
            try:
                minimum = hyperparameters["minimum"]
            except KeyError:
                raise KeyError(
                    "Cannot evaluate the probability when minimum is not given"
                )

        maximum = self["maximum"]
        if not self.fixed["maximum"]:
            try:
                maximum = hyperparameters["maximum"]
            except KeyError:
                raise KeyError(
                    "Cannot evaluate the probability when maximum is not given"
                )

        samples = bilby.core.prior.PowerLaw(alpha, minimum, maximum).sample(size=size)

        if size == 1:
            sample = samples[0]
        else:
            sample = samples

        return sample


class DeltaFunctionDistribution(BaseDistribution):
    """
    A distribution defining a delta function (useful if wanting to fix a
    parameter at a specific value if creating signals, or use as a null model).

    Parameters
    ----------
    name: str
        See :class:`~cwinpy.hierarchical.BaseDistribution`
    peak: float
        The value at which the delta function is non-zero.
    """

    def __init__(self, name, peak):
        # initialise
        super().__init__(name, "deltafunction", hyperparameters=dict(peak=peak))

    def log_pdf(self, value, hyperparameters={}):
        """
        The natural logarithm of the pdf of a delta function distribution.

        Parameters
        ----------
        value: float
            The value at which the probability is to be evaluated.
        hyperparameters: dict
            A dictionary containing the current values of the hyperparameters
            that need to be inferred.

        Returns
        -------
        logpdf:
            The natural logarithm of the probability at the given value.
        """

        peak = self["peak"]
        if not self.fixed["peak"]:
            try:
                peak = hyperparameters["peak"]
            except KeyError:
                raise KeyError("Cannot evaluate the probability when peak is not given")

        if value != peak:
            return -np.inf
        return 0.0

    def sample(self, hyperparameters={}, size=1):
        """
        Return the position of the delta function.

        Parameters
        ----------
        hyperparameters: dict
            A dictionary of the hyperparameter values (``peak``) that define
            the current state of the distribution.
        size: int
            The number of samples to draw from the distribution.

        Returns
        -------
        sample:
            A sample, or set of samples, from the distribution.
        """

        peak = self["peak"]
        if not self.fixed["peak"]:
            try:
                peak = hyperparameters["peak"]
            except KeyError:
                raise KeyError("Cannot evaluate the probability when peak is not given")

        if size == 1:
            return peak
        else:
            return peak * np.ones(size)


def create_distribution(name, distribution, distkwargs={}):
    """
    Function to create a distribution.

    An example of creating an exponential distribution, with a half-Gaussian
    prior on the mean would be:

    >>> from bilby.core.prior import HalfGaussian
    >>> sigma = 1e34  # width of half-Gaussian prior on mu
    >>> distkwargs = {"mu": HalfGaussian(name="mu", sigma=sigma)}
    >>> expdist = create_distribution("q22", "exponential", distkwargs)

    An example of creating a bimodal-Gaussian distribution, with modes fixed at
    particular values, fixed weights, but log-uniform priors on the mode
    widths, would be:

    >>> from bilby.core.prior import LogUniform
    >>> min = 1e28  # minimum of the width prior
    >>> max = 1e38  # maximum of the width prior
    >>> modes = [0.0, 1e32]  # fixed modes
    >>> weights = [0.7, 0.3]  # fixed weights
    >>> sigmas = [
    >>>     LogUniform(name="sigma0", minimum=min, maximum=max),
    >>>     LogUniform(name="sigma1", minimum=min, maximum=max),
    >>> ]
    >>> distkwargs = {
    >>>     "mu": modes,  # set "mu" for the modes
    >>>     "sigma": sigmas,  # set "sigma" for the widths
    >>>     "weight": weights,  # set "weight" for the weights
    >>> }
    >>> gaussdist = create_distribution("q22", "gaussian", distkwargs)

    Parameters
    ----------
    name: str
        The name of the parameter which the distribution represents.
    distribution: str, :class:`cwinpy.hierarchical.BaseDistribution`
        A string giving a valid distribution name. This is the distribution for
        which the hyperparameters are going to be inferred. If using a string,
        the distribution keyword arguments must be passed using ``distkwargs``.
    distkwargs: dict
        A dictionary of keyword arguments for the distribution that is being
        inferred.

    Returns
    -------
    distribution
        The distribution class.

    """

    if isinstance(distribution, BaseDistribution):
        return distribution
    elif isinstance(distribution, str):
        if distribution.lower() not in DISTRIBUTION_REQUIREMENTS.keys():
            raise ValueError('Unknown distribution type "{}"'.format(distribution))

        if distribution.lower() == "gaussian":
            return BoundedGaussianDistribution(name, **distkwargs)
        elif distribution.lower() == "exponential":
            return ExponentialDistribution(name, **distkwargs)
        elif distribution.lower() == "deltafunction":
            return DeltaFunctionDistribution(name, **distkwargs)
        elif distribution.lower() == "powerlaw":
            return PowerLawDistribution(name, **distkwargs)
    else:
        raise TypeError("Unknown distribution")


class MassQuadrupoleDistribution(object):
    """
    A class to infer the hyperparameters of the :math:`l=m=2` mass quadrupole
    distribution (or fiducial ellipticity :math:`\\varepsilon`) for a given
    selection of known pulsars (see, for example, [1]_).

    The class currently can attempt to fit the hyperparameters for the
    following distributions:

    * a :math:`n`-mode bounded Gaussian distribution defined by either fixed or
      unknown means, standard deviations and weights;
    * an exponential distribution defined by an unknown mean.
    * a power law distribution defined by an unknown power law index and fixed
      or unknown bounds.

    All distributions do not allow the quadrupole value to become negative.

    Parameters
    ----------
    data: :class:`bilby.core.result.ResultList`
        A :class:`bilby.core.result.ResultList` of outputs from running source
        parameter estimation using bilby for a set of individual CW sources.
        These can be from MCMC or nested sampler runs, but only the latter can
        be used if requiring a properly normalised evidence value.
    gridrange: array_like
        A list of values at which the :math:`Q_{22}` parameter posteriors
        should be interpolated, or a lower and upper bound in the range of
        values, which will be split into ``bins`` points spaced linearly in
        log-space. If not supplied this will instead be set using the posterior
        samples, with a minimum value at zero and a maximum given by the
        maximum of all posterior samples.
    bins: int
        The number of bins at which the posterior will be interpolated.
    distribution: :class:`cwinpy.hierarchical.BaseDistribution`, str
        A predefined distribution, or string giving a valid distribution name.
        This is the distribution for which the hyperparameters are going to be
        inferred. If using a string, the distribution keyword arguments must be
        passed using ``distkwargs``.
    distkwargs: dict
        A dictionary of keyword arguments for the distribution that is being
        inferred.
    bw: str, scalar, callable
        See the ``bw_method`` argument for :class:`scipy.stats.gaussian_kde`.
    sampler: str
        The name of the stochastic sampler method used by ``bilby`` for
        sampling the posterior. This defaults to use 'dynesty'.
    sampler_kwargs: dict
        A dictionary of arguments required by the given sampler.
    grid: dict
        A dictionary of values that define a grid in the parameter and
        hyperparameter space that can be used by a
        :class:`bilby.core.grid.Grid`. If given sampling will be performed on
        the grid, rather than using the stochastic sampler.
    integration_method: str
        The method to use for integration over the :math:`Q_{22}` parameter for
        each source. Default is 'numerical' to perform trapezium rule
        integration. Other allowed values are: 'sample' to sample over each
        individual :math:`Q_{22}` parameter for each source; or, 'expectation',
        which uses the :math:`Q_{22}` posterior samples to approximate the
        expectation value of the hyperparameter distribution. At the moment,
        these two additional methods may not be correct/reliable.
    nsamples: int
        This sets the number of posterior samples to store for either
        estimating KDEs or calculating expectation values from those passed in
        the data. This allows downsampling of large numbers of samples by
        randomly drawing a subsection of samples. If the number given is larger
        than the total number of samples for a given pulsar, then all samples
        will be used in that case. The default will be to use all samples, but
        this may lead to memory issues when using large numbers of pulsars.
    use_ellipticity: bool
        If True, work with fiducial ellipticity :math:`\\varepsilon` rather
        than mass quadrupole.

    To do
    -----

    Distributions that could be added include:

    * a Student's t-distributions with unknown mean and number of degrees of
      freedom.

    References
    ----------

    .. [1] M. Pitkin, C. Messenger & X. Fan, Phys. Rev. D, 98, 063001, 2018
       (`arXiv:1807.06726 <https://arxiv.org/abs/1807.06726>`_)
    """

    def __init__(
        self,
        data=None,
        gridrange=None,
        bins=100,
        distribution=None,
        distkwargs=None,
        bw="scott",
        sampler="dynesty",
        sampler_kwargs={},
        grid=None,
        integration_method="numerical",
        nsamples=None,
        use_ellipticity=False,
    ):
        self._posterior_samples = []
        self._pulsar_priors = []
        self._log_evidence = []
        self._likelihood_kdes_interp = []

        # set whether to use ellipticity rather than mass quadrupole
        self.use_ellipticity = use_ellipticity

        # set the values of q22/ellipticity at which to calculate the KDE
        # interpolator
        self.set_range(gridrange, bins)

        # set integration method
        self.set_integration_method(integration_method)

        # set the data
        self.add_data(data, bw=bw, nsamples=nsamples)

        # set the sampler
        if grid is None:
            self.set_sampler(sampler, sampler_kwargs)
        else:
            self.set_grid(grid)

        # set the distribution
        self.set_distribution(distribution, distkwargs)

    def set_range(self, gridrange, bins=100, prependzero=True):
        """
        Set the values of :math:`Q_{22}`, or ellipticity :math:`\\varepsilon`,
        either directly, or as a set of points linear in log-space defined by
        a lower and upper bounds and number of bins, at which to evaluate the
        posterior samples via their KDE to make an interpolator.

        Parameters
        ----------
        gridrange: array_like
            If this array contains two values it is assumed that these are the
            lower and upper bounds of a range, and the ``bins`` parameter sets
            the number of bins in log-space that the range will be split into.
            Otherwise, if more than two values are given it is assumed these
            are the values for :math:`Q_{22}` or :math:`\\varepsilon`.
        bins: int
            The number of bins the range is split into.
        prependzero: bool
            If setting an upper and lower range, this will prepend zero at the
            start of the range. Default is True.
        """

        self._bins = bins

        if gridrange is None:
            self._grid_interp_values = None
            return

        if len(gridrange) == 2:
            if gridrange[1] < gridrange[0]:
                raise ValueError("Grid range is badly defined")
            self._grid_interp_values = np.logspace(
                np.log10(gridrange[0]), np.log10(gridrange[1]), self._bins
            )

            if prependzero:
                self._grid_interp_values = np.insert(self._grid_interp_values, 0, 0)
        elif len(gridrange) > 2:
            self._grid_interp_values = gridrange
        else:
            raise ValueError("Grid range is badly defined")

    @property
    def interpolated_log_kdes(self):
        """
        Return the list of interpolation functions for the natural logarithm of
        the :math:`Q_{22}` likelihood functions after a Gaussian KDE has been
        applied.
        """

        return self._likelihood_kdes_interp

    def add_data(self, data, bw="scott", nsamples=None):
        """
        Set the data, i.e., the individual source posterior distributions, on
        which the hierarchical analysis will be performed.

        The posterior samples must include the ``Q22`` :math:`l=m=2` parameter,
        or the fiducial ellipticity parameter ``ELL``, for this inference to be
        performed.

        If using the "numerical" integration method, upon running the
        :meth:`~cwinpy.hierarchical.MassQuadrupoleDistribution.sample` method,
        these samples will be converted to a KDE (reflected about zero
        to avoid edge effects, and re-normalised), using
        :class:`scipy.stats.gaussian_kde`, which will be used as the
        data for hierarchical inference. If the posterior
        samples come with a Bayesian evidence value, and the prior is present,
        then these are used to convert the posterior distribution into a
        likelihood, which is what is then stored in the interpolation function.

        If using the "expectation" integration method, and if the posterior
        samples were not estimated using a uniform prior on ``Q22``/``ELL``,
        then the samples will be resampled from a uniform prior to attempt to
        generate samples from the likelihood.

        Parameters
        ----------
        data: :class:`bilby.core.result.ResultList`
            A list, or single, results from bilby containing posterior samples
            for a set of sources, or individual source.
        bw: str, scale, callable
            The Gaussian KDE bandwidth calculation method as required by
            :class:`scipy.stats.gaussian_kde`. The default is the 'scott'
            method.
        nsamples: int
            This sets the number of posterior samples to store and use from
            those passed in the data. This allows downsampling of large numbers
            of samples by randomly drawing a subsection of samples. If the
            number given is larger than the total number of samples for a given
            pulsar, then all samples will be used in that case. The default
            will be to use all samples, but this may lead to memory issues when
            using large numbers of pulsars.
        """

        # check the data is a ResultList
        if not isinstance(data, bilby.core.result.ResultList):
            if isinstance(data, (bilby.core.result.Result, str)):
                # convert to a ResultList
                data = bilby.core.result.ResultList([data])
            elif isinstance(data, list):
                data = bilby.core.result.ResultList(data)
            elif data is None:
                return
            else:
                raise TypeError("Data is not a known type")

        self._bw = bw

        for result in data:
            # check all posteriors contain Q22 or ellipticity
            if (
                "Q22" not in result.posterior.columns
                and "q22" not in result.posterior.columns
            ):
                if (
                    "ELL" in result.posterior.columns
                    or "ell" in result.posterior.columns
                ):
                    priorkey = "ell" if "ell" in result.posterior.columns else "ELL"
                    if not self.use_ellipticity:
                        # convert ellipticity into q22
                        result.posterior["q22"] = ellipticity_to_q22(
                            result.posterior[priorkey]
                        )
                else:
                    raise RuntimeError("Results do not contain Q22")
            else:
                priorkey = "q22" if "q22" in result.posterior.columns else "Q22"
                if self.use_ellipticity:
                    result.posterior["ell"] = q22_to_ellipticity(
                        result.posterior[priorkey]
                    )

            self._pulsar_priors.append(result.priors[priorkey])

        if nsamples is not None:
            if not isinstance(nsamples, int):
                raise TypeError("nsamples must be a positive integer")
            elif nsamples < 1:
                raise ValueError("nsamples must be a positive integer")

        # set number of samples to use
        if not hasattr(self, "_nsamples") and nsamples is not None:
            self._nsamples = nsamples
            numsamps = nsamples
        elif hasattr(self, "_nsamples") and nsamples is None:
            numsamps = self._nsamples
        else:
            numsamps = nsamples

        # create KDEs/add samples
        iniidx = len(self._posterior_samples)
        for i, result in enumerate(data):
            if self.use_ellipticity:
                keystr = "ell" if "ell" in result.posterior.columns else "ELL"
            else:
                keystr = "q22" if "q22" in result.posterior.columns else "Q22"

            samples = result.posterior[keystr]

            # reweight samples back to equivalent likelihood samples if prior
            # on Q22/ELL for PE was not uniform
            prior = self._pulsar_priors[iniidx + i]
            if self._integration_method == "expectation" and not isinstance(
                prior, bilby.core.prior.Uniform
            ):
                # resample to uniform prior
                possamps = result.posterior[keystr]
                lnweights = prior.log_prob(possamps)
                weights = np.exp(lnweights - np.max(lnweights))
                samples = possamps[weights > np.random.rand(len(weights))]

            # append samples
            if numsamps is None:
                self._posterior_samples.append(np.array(samples))
            else:
                if len(samples) < numsamps:
                    self._posterior_samples.append(np.array(samples))
                else:
                    # generate random choice of samples to store
                    sidx = np.random.default_rng().choice(
                        len(samples), numsamps, replace=False
                    )
                    self._posterior_samples.append(np.array(samples)[sidx])

            self._log_evidence.append(result.log_evidence)

    def set_distribution(self, distribution, distkwargs={}):
        """
        Set the distribution who's hyperparameters are going to be inferred.

        Parameters
        ----------
        distribution: :class:`cwinpy.hierarchical.BaseDistribution`, str
            A predefined distribution, or string giving a valid distribution
            name. If using a string, the distribution keyword arguments must be
            passed using ``distkwargs``.
        distkwargs: dict
            A dictionary of keyword arguments for the distribution that is being
            inferred.
        """

        self._distribution = None
        self._prior = None
        self._likelihood = None

        if distribution is None:
            return

        if isinstance(distribution, BaseDistribution):
            if self.use_ellipticity:
                if distribution.name.upper() != "ELL":
                    raise ValueError("Distribution name must be 'ELL'")
                else:
                    self._distribution = distribution
            else:
                if distribution.name.upper() != "Q22":
                    raise ValueError("Distribution name must be 'Q22'")
                else:
                    self._distribution = distribution
        elif isinstance(distribution, str):
            if self.use_ellipticity:
                self._distribution = create_distribution(
                    "ELL", distribution.lower(), distkwargs
                )
            else:
                self._distribution = create_distribution(
                    "Q22", distribution.lower(), distkwargs
                )

        # set the priors from the distribution
        self._set_priors()

    def _set_priors(self):
        """
        Set the priors based on those supplied via the distribution class.
        """

        # get the priors from the distribution
        if len(self._distribution.unknown_parameters) < 1:
            raise ValueError("Distribution has no parameters to infer")

        # add priors as PriorDict
        self._prior = bilby.core.prior.ConditionalPriorDict()

        for param, prior in zip(
            self._distribution.unknown_parameters, self._distribution.unknown_priors
        ):
            if isinstance(prior, bilby.core.prior.PriorDict):
                self._prior.update(prior)
            else:
                self._prior[param] = prior

    def _set_likelihood(self):
        """
        Set the likelihood.
        """

        samples = None
        grid = None
        likelihoods = None

        # set the grid
        if self._integration_method == "expectation":
            samples = self._posterior_samples
        else:
            if self._grid_interp_values is None:
                # set parameter range from data
                minmax = [np.inf, -np.inf]
                for psamples in self._posterior_samples:
                    minval = psamples.min()
                    maxval = psamples.max()
                    if minval < minmax[0]:
                        minmax[0] = minval
                    if maxval > minmax[1]:
                        minmax[1] = maxval

                self.set_range(minmax, self._bins)

            grid = self._grid_interp_values

            # generate KDEs from samples and create spline interpolants
            nkdes = len(self._likelihood_kdes_interp)
            if len(self._posterior_samples) > nkdes:
                for i in range(nkdes, len(self._posterior_samples)):
                    psamples = self._posterior_samples[i]
                    try:
                        # get reflected samples
                        samps = np.concatenate((psamples, -psamples))

                        # calculate KDE
                        kde = gaussian_kde(samps, bw_method=self._bw)

                        # use log pdf for the kde
                        interpvals = kde.logpdf(self._grid_interp_values) + np.log(
                            2.0
                        )  # multiply by 2 so pdf normalises to 1
                    except Exception as e:
                        raise RuntimeError("Problem creating KDE: {}".format(e))

                    # convert posterior to likelihood (if possible)
                    if np.isfinite(self._log_evidence[i]):
                        # multiply by evidence
                        interpvals += self._log_evidence[i]

                    # divide by prior
                    interpvals -= self._pulsar_priors[i].ln_prob(
                        self._grid_interp_values
                    )

                    # create and add interpolator (the tck tuple for a B-spline)
                    self._likelihood_kdes_interp.append(
                        splrep(self._grid_interp_values, interpvals)
                    )

            likelihoods = self._likelihood_kdes_interp

        self._likelihood = MassQuadrupoleDistributionLikelihood(
            self._distribution, likelihoods=likelihoods, samples=samples, grid=grid
        )

    def set_sampler(self, sampler="dynesty", sampler_kwargs={}):
        """
        Set the stochastic sampling method for ``bilby`` to use when sampling
        the parameter and hyperparameter posteriors.

        Parameters
        ----------
        sampler: str
            The name of the stochastic sampler method used by ``bilby`` for
            sampling the posterior. This defaults to use 'dynesty'.
        sampler_kwargs: dict
            A dictionary of arguments required by the given sampler.
        """

        self._sampler = sampler
        if self._sampler not in bilby.core.sampler.IMPLEMENTED_SAMPLERS:
            raise ValueError(
                'Sampler "{}" is not implemented in "bilby"'.format(self._sampler)
            )
        self._sampler_kwargs = sampler_kwargs
        self._use_grid = False  # set to not use the Grid sampling

    def set_grid(self, grid):
        """
        Set a grid on which to evaluate the hyperparameter posterior, as used
        by :class:`bilby.core.grid.Grid`.

        Parameters
        ----------
        grid: dict
            A dictionary of values that define a grid in the hyperparameter
            space that can be used by a :class:`bilby.core.grid.Grid` class.
        """

        if not isinstance(grid, dict):
            raise TypeError("Grid must be a dictionary")

        self._grid = grid
        self._use_grid = True

    def set_integration_method(self, integration_method="numerical"):
        """
        Set the method to use for integration over the :math:`Q_{22}` parameter
        for each source.

        Parameters
        ----------
        integration_method: str
            Default is 'numerical' to perform trapezium rule integration.
            The ther allowed value is 'expectation', which uses the
            :math:`Q_{22}` posterior samples to approximate the expectation
            value of the hyperparameter distribution.
        """

        if not isinstance(integration_method, str):
            raise TypeError("integration method must be a string")

        if integration_method.lower() not in ["numerical", "expectation"]:
            raise ValueError(
                "Unrecognised integration method type "
                "'{}'".format(integration_method)
            )

        self._integration_method = integration_method.lower()

    @property
    def result(self):
        """
        Return the ``bilby`` object containing the results. If evaluating the
        posterior over a grid this is a :class:`bilby.core.grid.Grid` object.
        If sampling using a stochastic sampler, this is a
        :class:`bilby.core.result.Result` object. If sampling has not yet been
        performed this returns ``None``.
        """

        if self._use_grid:
            if hasattr(self, "_grid_result"):
                return self._grid_result
            else:
                return None
        else:
            if hasattr(self, "_result"):
                return self._result
            else:
                return None

    def sample(self, **run_kwargs):
        """
        Sample the posterior distribution using ``bilby``. This can take
        keyword arguments required by the bilby
        :func:`~bilby.core.sampler.run_sampler` function.
        """

        # set up the likelihood function
        self._set_likelihood()

        # set use_ratio to False by default, i.e., don't use the likelihood
        # ratio
        run_kwargs.setdefault("use_ratio", False)

        if self._use_grid:
            self._grid_result = bilby.core.grid.Grid(
                self._likelihood, self._prior, grid_size=self._grid
            )
        else:
            self._result = bilby.run_sampler(
                likelihood=self._likelihood,
                priors=self._prior,
                sampler=self._sampler,
                **self._sampler_kwargs,
                **run_kwargs
            )

        return self.result

    def posterior_predictive(self, points, nsamples=100):
        """
        Return an iterator that will draw samples from the distribution
        hyperparameter posterior (once
        :meth:`~cwinpy.hierarchical.MassQuadrupoleDistribution.sample` has been
        run) and returns the associated distribution evaluated at a set of
        points.

        Currently this is only implemented to work using samples from a
        stochastic sampling method rather than posteriors evaluated on a grid.

        Parameters
        ----------
        points: array_like
            An array of Q22/ellipticity values at which to evaluate the
            distribution.
        nsamples: int
            The number of samples to draw from the distribution. This defaults
            to 100, but must be less than the number of posterior samples.
        """

        if self.result is None:
            raise RuntimeError("Sampling has not yet been performed")

        if self._use_grid:
            raise RuntimeError("Posterior predictive check can only use samples")

        # check grid
        if not isinstance(points, (tuple, list, np.ndarray)):
            raise TypeError("points must be array_like")

        if nsamples > len(self.result.posterior):
            raise ValueError(
                "Requested number of samples is greater than the number of posterior samples"
            )

        # chose random indexes of samples
        idx = np.random.choice(len(self.result.posterior), nsamples, replace=False)

        for i in idx:
            # get parameters of distribution for each sample
            hyper = {
                key: self.result.posterior[key][i]
                for key in self._distribution.unpacked_parameters
                if key in self.result.posterior.columns
            }

            # evaluate the distribution
            yield self._distribution.pdf(np.asarray(points), hyper)


class MassQuadrupoleDistributionLikelihood(bilby.core.likelihood.Likelihood):
    """
    The likelihood function for the inferring the hyperparameters of the
    mass quadrupole, :math:`Q_{22}`, distribution (or equivalently the
    fiducial ellipticity distribution).

    Parameters
    ----------
    distribution: :class:`cwinpy.hierarchical.BaseDistribution`
        The probability distribution for which the hyperparameters are going
        to be inferred.
    likelihoods: list
        A list of interpolation functions each of which gives the likelihood
        function for a single source.
    grid: array_like
        If given, the integration over the mass quadrupole distribution for
        each source is performed numerically on at these grid points. If not
        given, individual samples from :math:`Q_{22}` will be drawn from each
        source (i.e., equivalent to having a new :math:`Q_{22}` parameter for
        each source in the sampler).
    samples: list
        A list of arrays of :math:`Q_{22}` samples for each source. If this is
        given then these samples will be used to approximate the integral over
        independent :math:`Q_{22}` variables for each source.
    """

    def __init__(self, distribution, likelihoods=None, grid=None, samples=None):
        if not isinstance(distribution, BaseDistribution):
            raise TypeError("Distribution is not the correct type")

        # check that the distribution contains parameters to be inferred
        if len(distribution.unknown_parameters) < 1:
            raise ValueError("Distribution has no parameters to infer")

        inferred_parameters = {param: None for param in distribution.unknown_parameters}
        self.distribution = distribution
        self.grid = grid
        self.likelihoods = likelihoods
        self.samples = samples

        super().__init__(parameters=inferred_parameters)

    @property
    def likelihoods(self):
        return self._likelihoods

    @likelihoods.setter
    def likelihoods(self, like):
        if like is None:
            self._likelihoods = None
            self._nsources = 0
        elif not isinstance(like, list):
            raise TypeError("Likelihoods must be a list")
        else:
            if self.grid is not None:
                # evaluate the interpolated (log) likelihoods on the grid
                self._likelihoods = []
                for ll in like:
                    self._likelihoods.append(splev(self.grid, ll))
                self._nsources = len(like)
            else:
                raise ValueError("Grid must be set to evaluate likelihoods")

    @property
    def grid(self):
        return self._grid

    @grid.setter
    def grid(self, grid):
        if isinstance(grid, (list, np.ndarray)):
            self._grid = np.asarray(grid)
        elif grid is None:
            self._grid = None
        else:
            raise TypeError("Grid must be array-like")

    @property
    def samples(self):
        return self._samples

    @samples.setter
    def samples(self, samples):
        if samples is not None:
            if not isinstance(samples, (list, np.ndarray)):
                raise TypeError("samples value must be a list")

            if isinstance(samples, np.ndarray):
                if len(samples.shape) != 2:
                    raise ValueError("Samples must be a 2D array")

            for samplelist in samples:
                if not isinstance(samplelist, (list, np.ndarray)):
                    raise TypeError("Samples must be a list")

                if len(np.asarray(samplelist).shape) != 1:
                    raise ValueError("Source samples must be a 1d list")

            self._nsources = len(samples)

        self._samples = samples

    def log_likelihood(self):
        """
        Evaluate the log likelihood.
        """

        log_like = 0.0  # initialise the log likelihood

        if self.samples is not None:
            # log-likelihood using expectation value from samples
            for samps in self.samples:
                log_like += np.log(
                    np.mean(self.distribution.pdf(samps, self.parameters))
                )
        else:
            # evaluate the hyperparameter distribution
            logp = self.distribution.log_pdf(self.grid, self.parameters)

            # log-likelihood numerically integrating over grid
            for logl in self.likelihoods:
                log_like += logtrapz(logp + logl, self.grid, disable_checks=True)

        return log_like

    def noise_log_likelihood(self):
        """
        The log-likelihood for the unknown hyperparameters being equal to
        with zero.

        Note
        ----

        For distributions with hyperparameter priors that exclude zero this
        will always given :math:`-\\infty`.
        """

        for p in self.parameters:
            self.parameters[p] = 0.0

        return self.log_likelihood()

    def __len__(self):
        return self._nsources



#!/usr/bin/env python

"""
Run analysis to create two test pulsars and posterior samples to use
in the testing of the hierarchical code.
"""

import os
from collections import OrderedDict

import bilby
import numpy as np
from bilby.core.prior import PriorDict, Uniform
from cwinpy import HeterodynedData, TargetedPulsarLikelihood

# create a two fake pulsar parameter files
parcontent1 = """\
PSRJ     J0123+3456
RAJ      01:23:45.6789
DECJ     34:56:54.321
F0       567.89
F1       -1.2e-12
PEPOCH   56789
H0       0.0
COSIOTA  0.3
PSI      1.1
PHI0     2.4
DIST     2.0
"""

parcontent2 = """\
PSRJ     J2345-0123
RAJ      23:45:00.123
DECJ     -01:23:45.098
F0       212.09
F1       -6.7e-13
PEPOCH   56789
H0       0.0
COSIOTA  -0.4
PSI      0.1
PHI0     1.3
DIST     1.5
"""

label = "hierarchical_test_set"
outdir = "data"

if not os.path.isdir(outdir):
    os.makedirs(outdir)

detector = "H1"  # the detector to use
asd = [1e-24, 8e-25]  # noise amplitude spectral densities
times = np.linspace(1000000000.0, 1000086340.0, 1440)  # times
hets = []

# add content to the par file and create fake data
for i, content in enumerate([parcontent1, parcontent2]):
    parfile = os.path.join(outdir, "{}_{}.par".format(label, i))
    with open(parfile, "w") as fp:
        fp.write(content)

    # create some fake heterodyned data
    hets.append(
        HeterodynedData(times=times, par=parfile, fakeasd=asd[i], detector=detector)
    )

    # output the data
    hetfile = os.path.join(outdir, "{}_{}_data.txt".format(label, i))
    np.savetxt(
        hetfile,
        np.vstack((hets[i].times.value, hets[i].data.real, hets[i].data.imag)).T,
    )

# create priors
phi0range = [0.0, np.pi]
psirange = [0.0, np.pi / 2.0]
cosiotarange = [-1.0, 1.0]
q22range = [0.0, 1e38]
# h0range = [0., 1e-23]

# set prior for lalapps_pulsar_parameter_estimation_nested
priorfile = os.path.join(outdir, "{}_prior.txt".format(label))
priorcontent = """Q22 uniform {} {}
PHI0 uniform {} {}
PSI uniform {} {}
COSIOTA uniform {} {}
"""
with open(priorfile, "w") as fp:
    fp.write(priorcontent.format(*(q22range + phi0range + psirange + cosiotarange)))

# set prior for bilby
priors = OrderedDict()
priors["q22"] = Uniform(q22range[0], q22range[1], "q22", latex_label=r"$Q_{22}$")
priors["phi0"] = Uniform(
    phi0range[0], phi0range[1], "phi0", latex_label=r"$\phi_0$", unit="rad"
)
priors["psi"] = Uniform(
    psirange[0], psirange[1], "psi", latex_label=r"$\psi$", unit="rad"
)
priors["cosiota"] = Uniform(
    cosiotarange[0], cosiotarange[1], "cosiota", latex_label=r"$\cos{\iota}$"
)

Nlive = 1024  # number of nested sampling live points

# run bilby
for i, het in enumerate(hets):
    # set the likelihood for bilby
    likelihood = TargetedPulsarLikelihood(het, PriorDict(priors))

    thislabel = "{}_{}".format(label, i)

    # run bilby
    result = bilby.run_sampler(
        likelihood=likelihood,
        priors=priors,
        sampler="cpnest",
        nlive=Nlive,
        outdir=outdir,
        label=thislabel,
        use_ratio=False,
    )



#!/usr/bin/env python

"""
Compare cwinpy with lalapps_pulsar_parameter_estimation_nested for O1 data
from a single detector (H1). This uses data for the Crab pulsar, but restricts
the prior ranges on iota and psi based on observational constraints from the
pulsar wind nebula.
"""

import os
import subprocess as sp
from collections import OrderedDict

import corner
import h5py
import matplotlib.font_manager as font_manager
import numpy as np
from astropy.utils.data import download_file
from bilby.core.prior import (
    Gaussian,
    MultivariateGaussian,
    MultivariateGaussianDist,
    Uniform,
)
from comparitors import comparisons
from cwinpy.pe import pe
from lalinference import LALInferenceHDF5PosteriorSamplesDatasetName
from lalinference.io import read_samples
from matplotlib.lines import Line2D

# URL for ephemeris files
DOWNLOAD_URL = "https://git.ligo.org/lscsoft/lalsuite/raw/master/lalpulsar/lib/{}"

label = "single_detector_O1_data_restricted_prior"
outdir = "outputs"

if not os.path.isdir(outdir):
    os.makedirs(outdir)

# set the par file
parfile = os.path.join("data", "J0534+2200.par")

# set the data file
hetfile = os.path.join("data", "O1_Crab_H1.txt.gz")

# set the detector name
detector = "H1"

# create priors
phi0range = [0.0, np.pi]
psimeansigma = [2.1844, 1.6e-3]  # mean and standard deviation of Gaussian prior
h0range = [0.0, 1e-23]

# set prior for lalapps_pulsar_parameter_estimation_nested
priorfile = os.path.join(outdir, "{}_prior.txt".format(label))
priorcontent = """H0 uniform {} {}
PHI0 uniform {} {}
PSI gaussian {} {}
"""

# bi-model Gaussian distribution for ioa
iotamodes = 2  # 2 modes
iotameans = [1.085, 2.0566]
iotavars = [0.00022201, 0.00022201]
iotaweights = [1, 1]
iotaprior = "IOTA gmm {} [[{}],[{}]] [[[{}]],[[{}]]] [{},{}]\n"

with open(priorfile, "w") as fp:
    fp.write(priorcontent.format(*(h0range + phi0range + psimeansigma)))
    fp.write(iotaprior.format(*([iotamodes] + iotameans + iotavars + iotaweights)))

# set prior for bilby
priors = OrderedDict()
priors["h0"] = Uniform(h0range[0], h0range[1], "h0", latex_label=r"$h_0$")
priors["phi0"] = Uniform(
    phi0range[0], phi0range[1], "phi0", latex_label=r"$\phi_0$", unit="rad"
)
priors["psi"] = Gaussian(
    psimeansigma[0], psimeansigma[1], "psi", latex_label=r"$\psi$", unit="rad"
)

iotadist = MultivariateGaussianDist(
    names=["iota"],
    nmodes=iotamodes,
    mus=[[iotameans[0]], [iotameans[1]]],
    covs=[[[iotavars[0]]], [[iotavars[1]]]],
    weights=iotaweights,
)

priors["iota"] = MultivariateGaussian(
    iotadist, "iota", latex_label=r"$\iota$", unit="rad"
)

# run lalapps_pulsar_parameter_estimation_nested
try:
    execpath = os.environ["CONDA_PREFIX"]
except KeyError:
    raise KeyError(
        "Please work in a conda environment with lalsuite and cwinpy installed"
    )

execpath = os.path.join(execpath, "bin")

lppen = os.path.join(execpath, "lalapps_pulsar_parameter_estimation_nested")
n2p = os.path.join(execpath, "lalinference_nest2pos")

Nlive = 1024  # number of nested sampling live points
Nmcmcinitial = 0  # set to 0 so that prior samples are not resampled

outfile = os.path.join(outdir, "{}_nest.hdf".format(label))

# set ephemeris files
efile = download_file(DOWNLOAD_URL.format("earth00-40-DE405.dat.gz"), cache=True)
sfile = download_file(DOWNLOAD_URL.format("sun00-40-DE405.dat.gz"), cache=True)
tfile = download_file(DOWNLOAD_URL.format("te405_2000-2040.dat.gz"), cache=True)

# set the command line arguments
runcmd = " ".join(
    [
        lppen,
        "--verbose",
        "--input-files",
        hetfile,
        "--detectors",
        detector,
        "--par-file",
        parfile,
        "--prior-file",
        priorfile,
        "--Nlive",
        "{}".format(Nlive),
        "--Nmcmcinitial",
        "{}".format(Nmcmcinitial),
        "--outfile",
        outfile,
        "--ephem-earth",
        efile,
        "--ephem-sun",
        sfile,
        "--ephem-timecorr",
        tfile,
    ]
)

with sp.Popen(
    runcmd,
    stdout=sp.PIPE,
    stderr=sp.PIPE,
    shell=True,
    bufsize=1,
    universal_newlines=True,
) as p:
    for line in p.stderr:
        print(line, end="")

# convert nested samples to posterior samples
outpost = os.path.join(outdir, "{}_post.hdf".format(label))
runcmd = " ".join([n2p, "-p", outpost, outfile])
with sp.Popen(
    runcmd,
    stdout=sp.PIPE,
    stderr=sp.PIPE,
    shell=True,
    bufsize=1,
    universal_newlines=True,
) as p:
    for line in p.stdout:
        print(line, end="")

# get posterior samples
post = read_samples(outpost, tablename=LALInferenceHDF5PosteriorSamplesDatasetName)
lp = len(post["H0"])
postsamples = np.zeros((lp, len(priors)))
for i, p in enumerate(priors.keys()):
    postsamples[:, i] = post[p.upper()]

# get evidence
hdf = h5py.File(outpost, "r")
a = hdf["lalinference"]["lalinference_nest"]
evsig = a.attrs["log_evidence"]
evnoise = a.attrs["log_noise_evidence"]
hdf.close()

# run bilby via the pe interface
runner = pe(
    data_file=hetfile,
    par_file=parfile,
    prior=priors,
    detector=detector,
    sampler="dynesty",
    sampler_kwargs={"Nlive": Nlive, "walks": 40, "use_ratio": True},
    outdir=outdir,
    label=label,
    numba=True,
)

result = runner.result

# evaluate the likelihood on a grid
gridpoints = 35
grid_size = dict()
for p in priors.keys():
    grid_size[p] = np.linspace(
        np.min(result.posterior[p]), np.max(result.posterior[p]), gridpoints
    )

grunner = pe(
    data_file=hetfile,
    par_file=parfile,
    prior=priors,
    detector=detector,
    outdir=outdir,
    label=label,
    grid=True,
    grid_kwargs={"grid_size": grid_size},
    numba=True,
)

grid = grunner.grid

# output comparisons
comparisons(label, outdir, grid, priors, cred=0.9)

# plot results
fig = result.plot_corner(save=False, parameters=list(priors.keys()), color="b")
fig = corner.corner(
    postsamples,
    fig=fig,
    color="r",
    bins=50,
    smooth=0.9,
    quantiles=[0.16, 0.84],
    levels=(1 - np.exp(-0.5), 1 - np.exp(-2), 1 - np.exp(-9 / 2.0)),
    fill_contours=True,
    hist_kwargs={"density": True},
)
axes = fig.get_axes()
axidx = 0
for p in priors.keys():
    axes[axidx].plot(
        grid.sample_points[p],
        np.exp(grid.marginalize_ln_posterior(not_parameters=p) - grid.log_evidence),
        "k--",
    )
    axidx += 5

# custom legend
legend_elements = [
    Line2D([], [], color="r", label="lalapps_pulsar_parameter_estimation_nested"),
    Line2D([], [], color="b", label="cwinpy_pe"),
    Line2D([], [], color="k", ls="--", label="cwinpy_pe (grid)"),
]
font = font_manager.FontProperties(family="monospace")
leg = axes[3].legend(
    handles=legend_elements, loc="upper right", frameon=False, prop=font, handlelength=3
)
for line in leg.get_lines():
    line.set_linewidth(1.0)

fig.savefig(os.path.join(outdir, "{}_corner.png".format(label)), dpi=150)
